import unittest

import numpy as np

import qp
from qp.quantile_pdf_constructors import AbstractQuantilePdfConstructor, \
    PiecewiseLinear


class PiecewiseLinearTestCase(unittest.TestCase):
    """Tests for the CDF Spline Derivative PDF constructor for quantile parameterization."""

    def setUp(self):
        self.single_norm = qp.stats.norm(loc=3, scale=0.5)
        self.many_norm = qp.stats.norm(loc=np.array([[1], [2.5], [3]]), scale=np.array([[0.25], [0.5], [0.1]]))

        self.user_defined_quantiles = np.linspace(0.001, 0.999, 16)
        self.user_defined_grid = np.linspace(0, 5, 100)
        self.pdf_constructor = PiecewiseLinear

    def test_instantiation(self):
        """Base case make sure that we can instantiate the class"""
        user_defined_locations = self.single_norm.ppf(self.user_defined_quantiles)
        pdf_constructor = self.pdf_constructor(quantiles=self.user_defined_quantiles, locations=user_defined_locations)
        assert isinstance(pdf_constructor, AbstractQuantilePdfConstructor)

    def test_instantiation_for_multiple_distributions(self):
        """Base case make sure that we can instantiate a pdf reconstructor for multiple distributions"""
        user_defined_locations = self.many_norm.ppf(self.user_defined_quantiles)
        pdf_constructor = self.pdf_constructor(quantiles=self.user_defined_quantiles, locations=user_defined_locations)
        assert isinstance(pdf_constructor, AbstractQuantilePdfConstructor)

    def test_debug(self):
        """Ensure that debug returns expected values before `prepare_constructor` has been run"""
        user_defined_locations = self.single_norm.ppf(self.user_defined_quantiles)
        pdf_constructor = self.pdf_constructor(quantiles=self.user_defined_quantiles, locations=user_defined_locations)
        debug_quantiles, debug_locations, debug_first_derivatives, debug_adjusted_locations = pdf_constructor.debug()

        assert np.all(np.isclose(debug_quantiles, self.user_defined_quantiles))
        assert np.all(np.isclose(debug_locations, user_defined_locations))
        self.assertIsNone(debug_first_derivatives)
        self.assertIsNone(debug_adjusted_locations)

    def test_basic_construct_pdf(self):
        """Base case to ensure that `construct_pdf` method runs with minimum arguments for single distribution case
        Want to verify only that the machinery is working, and the result is not just zeros."""
        user_defined_locations = self.single_norm.ppf(self.user_defined_quantiles)
        pdf_constructor = self.pdf_constructor(quantiles=self.user_defined_quantiles, locations=user_defined_locations)
        results = pdf_constructor.construct_pdf(self.user_defined_grid)
        self.assertIsNot(np.sum(results), 0.0)

    def test_basic_construct_pdf_for_multiple_distributions(self):
        """Base case to ensure that `construct_pdf` method runs with minimum arguments for many-distribution case
        Want to verify only that the machinery is working, and the result is not just zeros."""
        user_defined_locations = self.many_norm.ppf(self.user_defined_quantiles)
        pdf_constructor = self.pdf_constructor(quantiles=self.user_defined_quantiles, locations=user_defined_locations)
        results = pdf_constructor.construct_pdf(self.user_defined_grid)
        self.assertIsNot(np.sum(results), 0.0)


"""This module implements a PDT distribution sub-class using interpolated grids
"""

import numpy as np

from scipy.stats import rv_continuous

from qp.pdf_gen import Pdf_rows_gen
from qp.conversion_funcs import extract_vals_at_x, extract_xy_vals, extract_xy_sparse
from qp.plotting import get_axes_and_xlims, plot_pdf_on_axes
from qp.utils import normalize_interp1d, interpolate_multi_x_y,\
    interpolate_multi_x_multi_y, interpolate_x_multi_y,\
    reshape_to_pdf_size
from qp.test_data import XBINS, XARRAY, YARRAY, TEST_XVALS
from qp.factory import add_class


class interp_gen(Pdf_rows_gen):
    """Interpolator based distribution

    Notes
    -----
    This implements a PDF using a set of interpolated values.

    This version use the same xvals for all the the PDFs, which
    allows for much faster evaluation, and reduces the memory
    usage by a factor of 2.

    The relevant data members are:

    xvals:  (n) x values

    yvals:  (npdf, n) y values

    Inside the range xvals[0], xvals[-1] tt simply takes a set of x and y values
    and uses `scipy.interpolate.interp1d` to build the PDF.
    Outside the range xvals[0], xvals[-1] the pdf() will return 0.

    The cdf() is constructed by integrating analytically computing the cumulative
    sum at the xvals grid points and interpolating between them.
    This will give a slight discrepency with the true integral of the pdf(),
    bit is much, much faster to evaluate.
    Outside the range xvals[0], xvals[-1] the cdf() will return (0 or 1), respectively

    The ppf() is computed by inverting the cdf().
    ppf(0) will return xvals[0]
    ppf(1) will return xvals[-1]
    """
    # pylint: disable=protected-access

    name = 'interp'
    version = 0

    _support_mask = rv_continuous._support_mask

    def __init__(self, xvals, yvals, *args, **kwargs):
        """
        Create a new distribution by interpolating the given values

        Parameters
        ----------
        xvals : array_like
          The x-values used to do the interpolation
        yvals : array_like
          The y-values used to do the interpolation
        """
        if np.size(xvals) != np.shape(yvals)[-1]: # pragma: no cover
            raise ValueError("Shape of xbins in xvals (%s) != shape of xbins in yvals (%s)" %
                             (np.size(xvals), np.shape(yvals)[-1]))
        self._xvals = np.asarray(xvals)

        # Set support
        self._xmin = self._xvals[0]
        self._xmax = self._xvals[-1]
        kwargs['shape'] = np.shape(yvals)[:-1]

        self._yvals = reshape_to_pdf_size(yvals, -1)

        check_input = kwargs.pop('check_input', True)
        if check_input:
            self._compute_ycumul()
            self._yvals = (self._yvals.T / self._ycumul[:,-1]).T
            self._ycumul = (self._ycumul.T / self._ycumul[:,-1]).T
        else:  # pragma: no cover
            self._ycumul = None

        super().__init__(*args, **kwargs)
        self._addmetadata('xvals', self._xvals)
        self._addobjdata('yvals', self._yvals)

    def _compute_ycumul(self):
        copy_shape = np.array(self._yvals.shape)
        self._ycumul = np.ndarray(copy_shape)
        self._ycumul[:, 0] = 0.5 * self._yvals[:, 0] * (self._xvals[1] - self._xvals[0])
        self._ycumul[:, 1:] = np.cumsum((self._xvals[1:] - self._xvals[:-1]) *
                                        0.5 * np.add(self._yvals[:,1:],
                                                     self._yvals[:,:-1]), axis=1)

    @property
    def xvals(self):
        """Return the x-values used to do the interpolation"""
        return self._xvals

    @property
    def yvals(self):
        """Return the y-valus used to do the interpolation"""
        return self._yvals

    def _pdf(self, x, row):
        # pylint: disable=arguments-differ
        return interpolate_x_multi_y(x, row, self._xvals, self._yvals,
                                     bounds_error=False, fill_value=0.).ravel()

    def _cdf(self, x, row):
        # pylint: disable=arguments-differ
        if self._ycumul is None:  # pragma: no cover
            self._compute_ycumul()
        return interpolate_x_multi_y(x, row, self._xvals, self._ycumul,
                                     bounds_error=False, fill_value=(0.,1.)).ravel()

    def _ppf(self, x, row):
        # pylint: disable=arguments-differ
        if self._ycumul is None:  # pragma: no cover
            self._compute_ycumul()

        return interpolate_multi_x_y(x, row, self._ycumul, self._xvals,
                                     bounds_error=False, fill_value=(self._xmin, self._xmax)).ravel()

    def _munp(self, m, *args):
        """ compute moments """
        # pylint: disable=arguments-differ
        # Silence floating point warnings from integration.
        with np.errstate(all='ignore'):
            vals = self.custom_generic_moment(m)
        return vals

    def custom_generic_moment(self, m):
        """ Compute the mth moment """
        m = np.asarray(m)
        dx = self._xvals[1] - self._xvals[0]
        return np.sum(self._xvals**m * self._yvals, axis=1) * dx

    def _updated_ctor_param(self):
        """
        Set the bins as additional constructor argument
        """
        dct = super()._updated_ctor_param()
        dct['xvals'] = self._xvals
        dct['yvals'] = self._yvals
        return dct

    @classmethod
    def get_allocation_kwds(cls, npdf, **kwargs):
        """
        Return the keywords necessary to create an 'empty' hdf5 file with npdf entries
        for iterative file writeout.  We only need to allocate the objdata columns, as
        the metadata can be written when we finalize the file.
        Parameters
        ----------
        npdf: (int) number of *total* PDFs that will be written out
        kwargs: (dict) dictionary of kwargs needed to create the ensemble
        """
        if 'xvals' not in kwargs: #pragma: no cover
            raise ValueError("required argument xvals not included in kwargs")
        ngrid = np.shape(kwargs['xvals'])[-1]
        return dict(yvals=((npdf, ngrid), 'f4'))

    @classmethod
    def plot_native(cls, pdf, **kwargs):
        """Plot the PDF in a way that is particular to this type of distibution

        For a interpolated PDF this uses the interpolation points
        """
        axes, _, kw = get_axes_and_xlims(**kwargs)
        return plot_pdf_on_axes(axes, pdf, pdf.dist.xvals, **kw)

    @classmethod
    def add_mappings(cls):
        """
        Add this classes mappings to the conversion dictionary
        """
        cls._add_creation_method(cls.create, None)
        cls._add_extraction_method(extract_vals_at_x, None)


    @classmethod
    def make_test_data(cls):
        """ Make data for unit tests """
        cls.test_data = dict(interp=dict(gen_func=interp, ctor_data=dict(xvals=XBINS, yvals=YARRAY),\
                                         convert_data=dict(xvals=XBINS), test_xvals=TEST_XVALS))

interp = interp_gen.create


class interp_irregular_gen(Pdf_rows_gen):
    """Interpolator based distribution

    Notes
    -----
    This implements a PDF using a set of interpolated values.

    This version use the different xvals for each the the PDFs, which
    allows for more precision.

    The relevant data members are:

    xvals:  (npdf, n) x values

    yvals:  (npdf, n) y values

    Inside the range xvals[:,0], xvals[:,-1] tt simply takes a set of x and y values
    and uses `scipy.interpolate.interp1d` to build the PDF.
    Outside the range xvals[:,0], xvals[:,-1] the pdf() will return 0.

    The cdf() is constructed by integrating analytically computing the cumulative
    sum at the xvals grid points and interpolating between them.
    This will give a slight discrepency with the true integral of the pdf(),
    bit is much, much faster to evaluate.
    Outside the range xvals[:,0], xvals[:,-1] the cdf() will return (0 or 1), respectively

    The ppf() is computed by inverting the cdf().
    ppf(0) will return min(xvals)
    ppf(1) will return max(xvals)
    """
    # pylint: disable=protected-access

    name = 'interp_irregular'
    version = 0

    _support_mask = rv_continuous._support_mask

    def __init__(self, xvals, yvals, *args, **kwargs):
        """
        Create a new distribution by interpolating the given values

        Parameters
        ----------
        xvals : array_like
          The x-values used to do the interpolation
        yvals : array_like
          The y-values used to do the interpolation
        """
        if np.shape(xvals) != np.shape(yvals): # pragma: no cover
            raise ValueError("Shape of xvals (%s) != shape of yvals (%s)" %
                             (np.shape(xvals), np.shape(yvals)))
        self._xvals = reshape_to_pdf_size(xvals, -1)

        self._xmin = np.min(self._xvals)
        self._xmax = np.max(self._xvals)
        kwargs['shape'] = np.shape(xvals)[:-1]

        check_input = kwargs.pop('check_input', True)
        self._yvals = reshape_to_pdf_size(yvals, -1)
        if check_input:
            self._yvals = normalize_interp1d(self._xvals, self._yvals)
        self._ycumul = None
        super().__init__(*args, **kwargs)
        self._addobjdata('xvals', self._xvals)
        self._addobjdata('yvals', self._yvals)

    def _compute_ycumul(self):
        copy_shape = np.array(self._yvals.shape)
        self._ycumul = np.ndarray(copy_shape)
        self._ycumul[:,0] = 0.
        self._ycumul[:,1:] = np.cumsum(self._xvals[:,1:]*self._yvals[:,1:] - self._xvals[:,:-1]*self._yvals[:,1:], axis=1)


    @property
    def xvals(self):
        """Return the x-values used to do the interpolation"""
        return self._xvals

    @property
    def yvals(self):
        """Return the y-valus used to do the interpolation"""
        return self._yvals

    def _pdf(self, x, row):
        # pylint: disable=arguments-differ
        return interpolate_multi_x_multi_y(x, row, self._xvals, self._yvals,
                                           bounds_error=False, fill_value=0.).ravel()

    def _cdf(self, x, row):
        # pylint: disable=arguments-differ
        if self._ycumul is None:  # pragma: no cover
            self._compute_ycumul()
        return interpolate_multi_x_multi_y(x, row, self._xvals, self._ycumul,
                                           bounds_error=False, fill_value=(0., 1.)).ravel()


    def _ppf(self, x, row):
        # pylint: disable=arguments-differ
        if self._ycumul is None:  # pragma: no cover
            self._compute_ycumul()
        return interpolate_multi_x_multi_y(x, row, self._ycumul, self._xvals,
                                           bounds_error=False, fill_value=(self._xmin, self._xmax)).ravel()


    def _updated_ctor_param(self):
        """
        Set the bins as additional constructor argument
        """
        dct = super()._updated_ctor_param()
        dct['xvals'] = self._xvals
        dct['yvals'] = self._yvals
        return dct

    @classmethod
    def get_allocation_kwds(cls, npdf, **kwargs):
        """
        Return the keywords necessary to create an 'empty' hdf5 file with npdf entries
        for iterative file writeout.  We only need to allocate the objdata columns, as
        the metadata can be written when we finalize the file.
        Parameters
        ----------
        npdf: (int) number of *total* PDFs that will be written out
        kwargs: (dict) dictionary of kwargs needed to create the ensemble
        """
        if 'xvals' not in kwargs: #pragma: no cover
            raise ValueError("required argument xvals not included in kwargs")
        ngrid = np.shape(kwargs['xvals'])[-1]
        return dict(xvals=((npdf, ngrid), 'f4'), yvals=((npdf, ngrid), 'f4'))

    @classmethod
    def plot_native(cls, pdf, **kwargs):
        """Plot the PDF in a way that is particular to this type of distibution

        For a interpolated PDF this uses the interpolation points
        """
        axes, _, kw = get_axes_and_xlims(**kwargs)
        xvals_row = pdf.dist.xvals
        return plot_pdf_on_axes(axes, pdf, xvals_row, **kw)

    @classmethod
    def add_mappings(cls):
        """
        Add this classes mappings to the conversion dictionary
        """
        cls._add_creation_method(cls.create, None)
        cls._add_extraction_method(extract_xy_vals, None)
        cls._add_extraction_method(extract_xy_sparse, None)


    @classmethod
    def make_test_data(cls):
        """ Make data for unit tests """
        cls.test_data = dict(interp_irregular=dict(gen_func=interp_irregular, ctor_data=dict(xvals=XARRAY, yvals=YARRAY),\
                                                   convert_data=dict(xvals=XBINS), test_xvals=TEST_XVALS))


interp_irregular = interp_irregular_gen.create
add_class(interp_gen)
add_class(interp_irregular_gen)



"""This module implements a PDT distribution sub-class using a Gaussian mixture model
"""
import os
import sys
import numpy as np
from scipy.stats import rv_continuous
from scipy import integrate as sciint
from scipy import interpolate as sciinterp
from qp import sparse_rep
from qp.factory import add_class
from qp.interp_pdf import interp_gen
from qp.conversion_funcs import extract_sparse_from_xy
from qp.test_data import TEST_XVALS, NPDF

class sparse_gen(interp_gen):
    """Sparse based distribution. The final behavior is similar to interp_gen, but the constructor
    takes a sparse representation to build the interpolator.
    Attempt to inherit from interp_gen : this is failing

    Notes
    -----
    This implements a qp interface to the original code SparsePz from M. Carrasco-Kind.

    """
    # pylint: disable=protected-access


    name = 'sparse'
    version = 0

    _support_mask = rv_continuous._support_mask

    def __init__(self, xvals, mu, sig, dims, sparse_indices, *args, **kwargs):
        self.sparse_indices = sparse_indices
        self._xvals = xvals
        self.mu = mu
        self.sig = sig
        self.dims = dims
        cut = kwargs.pop('cut', 1.e-5)
        #recreate the basis array from the metadata
        sparse_meta = dict(xvals=xvals, mu=mu, sig=sig, dims=dims)
        A = sparse_rep.create_basis(sparse_meta, cut=cut)
        #decode the sparse indices into basis indices and weights
        basis_indices, weights = sparse_rep.decode_sparse_indices(sparse_indices)
        #retrieve the weighted array of basis functions for each object
        pdf_y = A[:, basis_indices] * weights
        #normalize and sum the weighted pdfs
        x = sparse_meta['xvals']
        y = pdf_y.sum(axis=-1)
        norms = sciint.trapz(y.T, x)
        y /= norms
        kwargs.setdefault('xvals', x)
        kwargs.setdefault('yvals', y.T)
        super().__init__(*args, **kwargs)

        self._clearobjdata()
        self._addmetadata('xvals', self._xvals)
        self._addmetadata('mu', self.mu)
        self._addmetadata('sig', self.sig)
        self._addmetadata('dims', self.dims)
        self._addobjdata('sparse_indices', self.sparse_indices)

    def _updated_ctor_param(self):
        """
        Add the two constructor's arguments for the Factory
        """
        dct = super()._updated_ctor_param()
        dct['sparse_indices'] = self.sparse_indices
        dct['xvals'] = self._xvals
        dct['mu'] = self.mu
        dct['sig'] = self.sig
        dct['dims'] = self.dims
        return dct

    @classmethod
    def get_allocation_kwds(cls, npdf, **kwargs):
        if 'dims' not in kwargs:
            raise ValueError("required argument dims not in kwargs") #pragma: no cover
        nsp = np.array(kwargs['dims']).flatten()[4]
        return dict(sparse_indices=((npdf, nsp), 'i8'))

    @classmethod
    def add_mappings(cls):
        """
        Add this classes mappings to the conversion dictionary
        """
        cls._add_creation_method(cls.create, None)
        cls._add_extraction_method(extract_sparse_from_xy, None)


    @staticmethod
    def build_test_data():
        """build a test case out of real pdfs"""
        qproot = sys.modules['qp'].__path__[0]
        filein = os.path.join(qproot, './data/CFHTLens_sample.P.npy')
        #FORMAT FILE, EACH ROW IS THE PDF FOR EACH GALAXY, LAST ROW IS THE REDSHIFT POSITION
        P = np.load(filein)
        z = P[-1]
        P = P[:NPDF]
        P = P / sciint.trapz(P, z).reshape(-1, 1)
        minz = np.min(z)
        nz = 301
        _, j = np.where(P > 0)
        maxz = np.max(z[j+1])
        newz = np.linspace(minz, maxz, nz)
        interp = sciinterp.interp1d(z, P, assume_sorted=True)
        newpdf = interp(newz)
        newpdf = newpdf / sciint.trapz(newpdf, newz).reshape(-1, 1)
        sparse_idx, meta, _ = sparse_rep.build_sparse_representation(newz, newpdf, verbose=False)
        return sparse_idx, meta


    @classmethod
    def make_test_data(cls):
        SPARSE_IDX, META = cls.build_test_data()

        cls.test_data = dict(sparse=dict(gen_func=sparse, \
                                         ctor_data=dict(xvals=META['xvals'], mu=META['mu'], sig=META['sig'],\
                                                        dims=META['dims'], sparse_indices=SPARSE_IDX),\
                                         test_xvals=TEST_XVALS), )


sparse = sparse_gen.create

add_class(sparse_gen)



"""
Unit tests for PDF class
"""
import copy
import logging
import numpy as np
import unittest
import qp
from qp import test_data
import os
import sys

from qp.test_funcs import assert_all_small, assert_all_close, build_ensemble
from qp.plotting import init_matplotlib
from qp.quantile_pdf_constructors import AbstractQuantilePdfConstructor


class EnsembleTestCase(unittest.TestCase):
    """ Class to test qp.Ensemble functionality """

    def setUp(self):
        """
        Make any objects that are used in multiple tests.
        """

    def tearDown(self):
        "Clean up any mock data files created by the tests."

    @staticmethod
    def _run_ensemble_funcs(ens, xpts):
        """Run the test for a practicular class"""

        pdfs = ens.pdf(xpts)
        cdfs = ens.cdf(xpts)
        logpdfs = ens.logpdf(xpts)
        logcdfs = ens.logcdf(xpts)

        if hasattr(ens.gen_obj, 'npdf'):
            assert ens.npdf == ens.gen_obj.npdf

        with np.errstate(all='ignore'):
            assert np.allclose(np.log(pdfs), logpdfs, atol=1e-9)
            assert np.allclose(np.log(cdfs), logcdfs, atol=1e-9)

        binw = xpts[1:] - xpts[0:-1]
        check_cdf = ((pdfs[:,0:-1] + pdfs[:,1:]) * binw /2).cumsum(axis=1) - cdfs[:,1:]
        assert_all_small(check_cdf, atol=5e-2, test_name="cdf")

        hist = ens.histogramize(xpts)[1]
        hist_check = ens.frozen.histogramize(xpts)[1]
        assert_all_small(hist-hist_check, atol=1e-5, test_name="hist")

        ppfs = ens.ppf(test_data.QUANTS)
        check_ppf = ens.cdf(ppfs) - test_data.QUANTS
        assert_all_small(check_ppf, atol=2e-2, test_name="ppf")

        sfs = ens.sf(xpts)
        check_sf = sfs + cdfs
        assert_all_small(check_sf-1, atol=2e-2, test_name="sf")

        _ = ens.isf(test_data.QUANTS)
        check_isf = ens.cdf(ppfs) + test_data.QUANTS[::-1]
        assert_all_small(check_isf-1, atol=2e-2, test_name="isf")

        samples = ens.rvs(size=1000)
        assert samples.shape[0] == ens.frozen.npdf
        assert samples.shape[1] == 1000

        median = ens.median()
        mean = ens.mean()        
        var = ens.var()
        std = ens.std()
        entropy = ens.entropy()

        _ = ens.stats()
        modes = ens.mode(xpts)

        assert median.size == ens.npdf
        assert mean.size == ens.npdf
        assert np.std(mean) > 1e-8
        assert var.size == ens.npdf
        assert std.size == ens.npdf
        assert entropy.size == ens.npdf
        assert modes.size == ens.npdf

        integral = ens.integrate(limits=(ens.gen_obj.a, ens.gen_obj.a))
        interval = ens.interval(0.05)

        assert integral.size == ens.npdf
        assert interval[0].size == ens.npdf

        for N in range(3):
            moment_partial = ens.moment_partial(N, limits=(test_data.XMIN, test_data.XMAX))
            calc_moment = qp.metrics.calculate_moment(ens, N, limits=(test_data.XMIN, test_data.XMAX))
            assert_all_close(moment_partial, calc_moment, rtol=5e-2, test_name="moment_partial_%i" % N)

            sps_moment = ens.moment(N)
            assert sps_moment.size == ens.npdf
            #assert_all_close(sps_moment.flatten(), moment_partial.flatten(), rtol=5e-2, test_name="moment_%i" % N)
            #pmf = ens.pmf(N)
            #logpmf = ens.logpmf(N)

        init_matplotlib()
        axes = ens.plot(xlim=(xpts[0], xpts[-1]))
        ens.plot_native(axes=axes)

        red_ens = ens[np.arange(5)]
        red_pdf = red_ens.pdf(xpts)

        check_red = red_pdf - pdfs[0:5]
        assert_all_small(check_red, atol=1e-5, test_name="red")

        if hasattr(ens.gen_obj, 'npdf'): # skip scipy norm
            commList = [None]
            try:
                import mpi4py.MPI
                commList.append(mpi4py.MPI.COMM_WORLD)
            except ImportError:
                pass
            for comm in commList:
                try:
                    group, fout = ens.initializeHdf5Write("testwrite.hdf5", ens.npdf, comm)
                except TypeError:
                    continue
                ens.writeHdf5Chunk(group, 0, ens.npdf)
                ens.finalizeHdf5Write(fout)
                readens = qp.read("testwrite.hdf5")
                assert readens.metadata().keys() == ens.metadata().keys()
                assert readens.objdata().keys() == ens.objdata().keys()
                os.remove("testwrite.hdf5")


    @staticmethod
    def _run_merge_tests(ens, xpts):
        npdf = ens.npdf
        pdf_orig = ens.pdf(xpts)

        ens_cat = qp.concatenate([ens, ens])
        ens.append(ens)

        pdf_cat = ens_cat.pdf(xpts)

        modes = np.array([xpts[idx] for idx in np.squeeze(np.argmax(pdf_cat, axis=1))])

        ens_cat.set_ancil({"mode":modes})
        pdf_app = ens.pdf(xpts)

        mask = np.concatenate([np.ones((npdf), 'bool'), np.zeros((npdf), 'bool')])
        ens_check = ens_cat[mask]
        pdf_check = ens_check.pdf(xpts)

        assert_all_close(pdf_cat, pdf_app, atol=5e-8, test_name="merge_1")
        assert_all_close(pdf_orig, pdf_check, atol=5e-8, test_name="merge_2")
        assert_all_close(ens_cat.ancil['mode'][mask], modes[mask], atol=5e-8, test_name="mode")


    def test_norm(self):
        """ Run the ensemble tests on an ensemble of scipy.stats.norm distributions """
        key = 'norm'
        cls_test_data = qp.stats.norm_gen.test_data[key]  #pylint: disable=no-member
        ens_norm = build_ensemble(cls_test_data)
        assert hasattr(ens_norm, 'gen_func')
        assert isinstance(ens_norm.gen_obj, qp.stats.norm_gen)  #pylint: disable=no-member
        assert 'loc' in ens_norm.frozen.kwds
        self._run_ensemble_funcs(ens_norm, cls_test_data['test_xvals'])
        self._run_merge_tests(ens_norm, cls_test_data['test_xvals'])


    def test_hist(self):
        """ Run the ensemble tests on an ensemble of qp.hist distributions """
        key = 'hist'
        qp.hist_gen.make_test_data()
        cls_test_data = qp.hist_gen.test_data[key]
        ens_h = build_ensemble(cls_test_data)
        assert isinstance(ens_h.gen_obj, qp.hist_gen)
        self._run_ensemble_funcs(ens_h, cls_test_data['test_xvals'])
        self._run_merge_tests(ens_h, cls_test_data['test_xvals'])

        pdfs_mod = copy.copy(ens_h.dist.pdfs)
        pdfs_mod[:,7] = 0.5*pdfs_mod[:,7]
        ens_h.update_objdata(dict(pdfs=pdfs_mod))

    def test_interp(self):
        """ Run the ensemble tests on an ensemble of qp.interp distributions """
        key = 'interp'
        qp.interp_gen.make_test_data()
        cls_test_data = qp.interp_gen.test_data[key]
        ens_i = build_ensemble(cls_test_data)
        assert isinstance(ens_i.gen_obj, qp.interp_gen)
        self._run_ensemble_funcs(ens_i, cls_test_data['test_xvals'])

    def test_packed_interp(self):
        """ Run the ensemble tests on an ensemble of qp.packed_interp distributions """
        key = 'lin_packed_interp'
        qp.packed_interp_gen.make_test_data()
        cls_test_data = qp.packed_interp_gen.test_data[key]
        ens_i = build_ensemble(cls_test_data)
        assert isinstance(ens_i.gen_obj, qp.packed_interp_gen)
        self._run_ensemble_funcs(ens_i, cls_test_data['test_xvals'])
        assert np.isfinite(ens_i.dist.yvals).all()

    def test_iterator(self):
        """ Test the iterated read """
        QP_DIR = os.path.abspath(os.path.dirname(qp.__file__))
        data_file = os.path.join(QP_DIR, 'data', 'test.hdf5')
        ens = qp.read(data_file)
        data_length = qp.data_length(data_file)
        assert data_length == ens.npdf
        itr = qp.iterator(data_file, 10)
        test_grid = np.linspace(0., 1., 11)
        for start, end, ens_i in itr:
            check_vals = ens[start:end].pdf(test_grid)
            test_vals = ens_i.pdf(test_grid)
            assert np.allclose(check_vals, test_vals)

    def test_quant_get_default_pdf_constructor_name(self):
        """Test that the getter for pdf constructor name works"""
        quantiles = np.linspace(0.001, 0.999, 16)
        locations = np.linspace(0, 5, 16)
        quant_dist = qp.quant(quants=quantiles, locs=locations)
        self.assertEqual(quant_dist.dist.pdf_constructor_name, 'piecewise_linear')

    def test_quant_get_default_pdf_constructor(self):
        """Test that the getter for pdf constructor returns an AbstractQuantilePdfConstructor"""
        quantiles = np.linspace(0.001, 0.999, 16)
        locations = np.linspace(0, 5, 16)
        quant_dist = qp.quant(quants=quantiles, locs=locations)
        assert isinstance(quant_dist.dist.pdf_constructor, AbstractQuantilePdfConstructor)

    def test_quant_change_pdf_constructor(self):
        """Test that changing the pdf constructor works as expected"""
        quantiles = np.linspace(0.001, 0.999, 16)
        locations = np.linspace(0, 5, 16)
        quant_dist = qp.quant(quants=quantiles, locs=locations)
        quant_dist.dist.pdf_constructor_name = 'piecewise_constant'
        self.assertEqual(quant_dist.dist.pdf_constructor_name, 'piecewise_constant')

    def test_quant_change_pdf_constructor_raises(self):
        """Verify that attempting to change the pdf constructor to one that
        isn't in the dictionary, will raise an error."""
        quantiles = np.linspace(0.001, 0.999, 16)
        locations = np.linspace(0, 5, 16)
        quant_dist = qp.quant(quants=quantiles, locs=locations)
        with self.assertRaises(ValueError):
            quant_dist.dist.pdf_constructor_name = 'drewtonian'

    def test_quant_change_pdf_constructor_warns(self):
        """Verify that attempting to change the pdf constructor to the one
        currently being used will log a warning."""
        quantiles = np.linspace(0.001, 0.999, 16)
        locations = np.linspace(0, 5, 16)
        quant_dist = qp.quant(quants=quantiles, locs=locations)
        with self.assertLogs(level=logging.WARNING) as log:
            quant_dist.dist.pdf_constructor_name = 'piecewise_linear'
            self.assertIn('Already using', log.output[0])

if __name__ == '__main__':
    unittest.main()



"""This small module collects inputs for unit tests"""

import os
import numpy as np
import scipy.stats as sps

np.random.seed(1234)

NPDF = 11
NBIN = 61
NSAMPLES = 100
XMIN = 0.
XMAX = 5.
LOC = np.expand_dims(np.linspace(0.5, 2.5, NPDF), -1)
SCALE = np.expand_dims(np.linspace(0.2, 1.2, NPDF), -1)
LOC_SHIFTED = LOC + SCALE
TEST_XVALS = np.linspace(XMIN, XMAX, 201)
XBINS = np.linspace(XMIN, XMAX, NBIN)
XARRAY = np.ones((NPDF, NBIN))*XBINS
YARRAY = np.expand_dims(np.linspace(0.5, 2.5, NPDF), -1)*(1. + 0.1*np.random.uniform(size=(NPDF, NBIN)))
HIST_DATA = YARRAY[:,0:-1]
QUANTS = np.linspace(0.01, 0.99, NBIN)
QLOCS = sps.norm(loc=LOC, scale=SCALE).ppf(QUANTS)
SAMPLES = sps.norm(loc=LOC, scale=SCALE).rvs(size=(NPDF, NSAMPLES))

MEAN_MIXMOD = np.vstack([np.linspace(0.5, 2.5, NPDF), np.linspace(0.5, 1.5, NPDF), np.linspace(1.5, 2.5, NPDF)]).T
STD_MIXMOD = np.vstack([np.linspace(0.2, 1.2, NPDF), np.linspace(0.2, 0.5, NPDF), np.linspace(0.2, 0.5, NPDF)]).T
WEIGHT_MIXMOD = np.vstack([0.7*np.ones((NPDF)), 0.2*np.ones((NPDF)), 0.1*np.ones((NPDF))]).T

HIST_TOL = 4./NBIN
QP_TOPDIR = os.path.dirname(os.path.dirname(__file__))



import sys
import os

# Provide path to the python modules we want to run autodoc on
sys.path.insert(0, os.path.abspath('../qp'))

import qp

# Avoid imports that may be unsatisfied when running sphinx, see:
# http://stackoverflow.com/questions/15889621/sphinx-how-to-exclude-imports-in-automodule#15912502
autodoc_mock_imports = ["scipy","scipy.interpolate", "sklearn"]

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.mathjax',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode']

# on_rtd is whether we are on readthedocs.org, this line of code grabbed from docs.readthedocs.org
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'

if not on_rtd:
    # only import and set the theme if we're building docs locally
    import sphinx_rtd_theme
    html_theme = 'sphinx_rtd_theme'
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# otherwise, readthedocs.org uses their theme by default, so
# no need to specify it.

master_doc = 'index'
autosummary_generate = True
autoclass_content = "class"
autodoc_default_flags = ["members", "no-special-members"]
autodoc_member_order = "bysource"


html_sidebars = { '**': ['globaltoc.html', 'relations.html', 'sourcelink.html', 'searchbox.html'], }

project = u'qp'
author = u'Alex Malz and Phil Marshall'
copyright = u'2016, ' + author
version = qp.__version__
release = qp.__version__



"""This module implements continous distributions generators that inherit from the
`scipy.stats.rv_continuous` class

If you would like to add a sub-class, please read the instructions on subclassing
here:
https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.rv_continuous.html


Open questions:
1) At this time the normalization is not enforced for many of the PDF types.  It is assumed that
the user values give correct normalization.  We should think about this more.

2) At this time for most of the distributions, only the _pdf function is overridden.  This is all that
is required to inherit from `scipy.stats.rv_continuous`; however, providing implementations of some of
_logpdf, _cdf, _logcdf, _ppf, _rvs, _isf, _sf, _logsf could speed the code up a lot in some cases.

"""
import sys

import numpy as np
from numpy import asarray

from scipy.stats import rv_continuous
from scipy.stats._distn_infrastructure import rv_continuous_frozen

from qp.utils import reshape_to_pdf_size, reshape_to_pdf_shape
from qp.dict_utils import get_val_or_default, set_val_or_default, pretty_print
from qp.plotting import plot_dist_pdf


class Pdf_gen:
    """Interface class to extend `scipy.stats.rv_continuous` with
    information needed for `qp`

    Notes
    -----
    Metadata are elements that are the same for all the PDFs
    These include the name and version of the PDF generation class, and
    possible data such as the bin edges used for histogram representations

    Object data are elements that differ for each PDFs
    """

    _reader_map = {}
    _creation_map = {}
    _extraction_map = {}

    def __init__(self, *args, **kwargs):
        """C'tor"""
        # pylint: disable=unused-argument
        self._metadata = {}
        self._objdata = {}
        self._addclassmetadata(type(self))

    def _addclassmetadata(self, cls):
        self._metadata['pdf_name'] = np.array([cls.name.encode()])
        self._metadata['pdf_version'] = np.array([cls.version])

    def _addmetadata(self, key, val):
        self._metadata[key] = np.expand_dims(val, 0)

    def _addobjdata(self, key, val):
        self._objdata[key] = val

    def _clearobjdata(self):
        self._objdata = {}

    @property
    def metadata(self):
        """Return the metadata for this set of PDFs"""
        return self._metadata

    @property
    def objdata(self):
        """Return the object data for this set of PDFs"""
        return self._objdata

    @classmethod
    def creation_method(cls, method=None):
        """Return the method used to create a PDF of this type"""
        return get_val_or_default(cls._creation_map, method)

    @classmethod
    def extraction_method(cls, method=None):
        """Return the method used to extract data to create a PDF of this type"""
        return get_val_or_default(cls._extraction_map, method)

    @classmethod
    def reader_method(cls, version=None):
        """Return the method used to convert data read from a file PDF of this type"""
        return get_val_or_default(cls._reader_map, version)

    @classmethod
    def add_method_dicts(cls):
        """Add empty method dicts"""
        cls._reader_map = {}
        cls._creation_map = {}
        cls._extraction_map = {}

    @classmethod
    def _add_creation_method(cls, the_func, method):
        """Add a method used to create a PDF of this type"""
        set_val_or_default(cls._creation_map, method, the_func)

    @classmethod
    def _add_extraction_method(cls, the_func, method):
        """Add a method used to extract data to create a PDF of this type"""
        set_val_or_default(cls._extraction_map, method, the_func)

    @classmethod
    def _add_reader_method(cls, the_func, version): #pragma: no cover
        """Add a method used to convert data read from a file PDF of this type"""
        set_val_or_default(cls._reader_map, version, the_func)

    @classmethod
    def print_method_maps(cls, stream=sys.stdout):
        """Print the maps showing the methods"""
        pretty_print(cls._creation_map, ["Create  "], stream=stream)
        pretty_print(cls._extraction_map, ["Extract "], stream=stream)
        pretty_print(cls._reader_map, ["Reader  "], stream=stream)


    @classmethod
    def create_gen(cls, **kwds):
        """Create and return a `scipy.stats.rv_continuous` object using the
        keyword arguemntets provided"""
        kwds_copy = kwds.copy()
        name = kwds_copy.pop('name', 'dist')
        return (cls(name=name), kwds_copy)

    @classmethod
    def create(cls, **kwds):
        """Create and return a `scipy.stats.rv_frozen` object using the
        keyword arguemntets provided"""
        # pylint: disable=not-callable
        obj, kwds_freeze = cls.create_gen(**kwds)
        return obj(**kwds_freeze)

    @classmethod
    def plot(cls, pdf, **kwargs):
        """Plot the pdf as a curve"""
        return plot_dist_pdf(pdf, **kwargs)

    @classmethod
    def plot_native(cls, pdf, **kwargs):
        """Plot the PDF in a way that is particular to this type of distibution

        This defaults to plotting it as a curve, but this can be overwritten
        """
        return plot_dist_pdf(pdf, **kwargs)

    @classmethod
    def get_allocation_kwds(cls, npdf, **kwargs):
        """Return kwds necessary to create 'empty' hdf5 file with npdf entries
        for iterative writeout
        """
        raise NotImplementedError()  #pragma: no cover


class rv_frozen_func(rv_continuous_frozen):
    """Trivial extention of `scipy.stats.rv_frozen`
    that includes the number of PDFs it represents
    """

    def __init__(self, dist, shape, *args, **kwds):
        """C'tor

        Parameters
        ----------
        dist : `scipy.stats.rv_continuous`
            The underlying distribution
        npdf : `int`
            The number of PDFs this object represents
        """
        self._shape = shape
        self._npdf = np.product(shape).astype(int)
        self._ndim = np.size(shape)
        super().__init__(dist, *args, **kwds)

    @property
    def ndim(self):
        """Return the number of dimensions of PDFs in this ensemble"""
        return self._ndim

    @property
    def shape(self):
        """Return the shape of the set of PDFs this object represents"""
        return self._shape

    @property
    def npdf(self):
        """Return the number of PDFs this object represents"""
        return self._npdf

    def histogramize(self, bins):
        """
        Computes integrated histogram bin values for all PDFs

        Parameters
        ----------
        bins: ndarray, float, optional
            Array of N+1 endpoints of N bins

        Returns
        -------
        self.histogram: ndarray, tuple, ndarray, floats
            Array of pairs of arrays of lengths (N+1, N) containing endpoints
            of bins and values in bins
        """
        cdf_vals = reshape_to_pdf_size(self.cdf(bins), -1)
        bin_vals = cdf_vals[:,1:] - cdf_vals[:,0:-1]
        return (bins, reshape_to_pdf_shape(bin_vals, self._shape, bins.size-1))


class rv_frozen_rows(rv_continuous_frozen):
    """Trivial extention of `scipy.stats.rv_frozen`
    that to use when we want to have a collection
    of distribution of objects such as histograms or splines,
    where each object represents a single distribtuion
    """

    def __init__(self, dist, shape, *args, **kwds):
        """C'tor"""
        self._shape = shape
        self._npdf = np.product(shape).astype(int)
        self._ndim = np.size(shape)
        if self._npdf is not None:
            kwds.setdefault('row', np.expand_dims(np.arange(self._npdf).reshape(self._shape), -1))
        super().__init__(dist, *args, **kwds)

    @property
    def ndim(self):
        """Return the number of dimensions of PDFs in this ensemble"""
        return self._ndim

    @property
    def shape(self):
        """Return the shape of the set of PDFs this object represents"""
        return self._shape

    @property
    def npdf(self):
        """Return the number of PDFs this object represents"""
        return self._npdf

    def histogramize(self, bins):
        """
        Computes integrated histogram bin values for all PDFs

        Parameters
        ----------
        bins: ndarray, float, optional
            Array of N+1 endpoints of N bins

        Returns
        -------
        self.histogram: ndarray, tuple, ndarray, floats
            Array of pairs of arrays of lengths (N+1, N) containing endpoints
            of bins and values in bins
        """
        cdf_vals = reshape_to_pdf_size(self.cdf(bins), -1)
        bin_vals = cdf_vals[:,1:] - cdf_vals[:,0:-1]
        return (bins, reshape_to_pdf_shape(bin_vals, self._shape, bins.size-1))



class Pdf_rows_gen(rv_continuous, Pdf_gen):
    """Class extend `scipy.stats.rv_continuous` with
    information needed for `qp` when we want to have a collection
    of distribution of objects such as histograms or splines,
    where each object represents a single distribtuion

    """
    def __init__(self, *args, **kwargs):
        """C'tor"""
        self._shape = kwargs.pop('shape', (1))
        self._npdf = np.product(self._shape).astype(int)
        super().__init__(*args, **kwargs)

    @property
    def shape(self):
        """Return the shape of the set of PDFs this object represents"""
        return self._shape

    @property
    def npdf(self):
        """Return the number of PDFs this object represents"""
        return self._npdf

    @staticmethod
    def _sliceargs(x, row, *args):  #pragma: no cover
        if np.size(x) == 1 or np.size(row) == 1:
            return False, x, row, args
        xx = np.unique(x)
        rr = np.unique(row)
        if np.size(xx) == np.size(x):
            xx = x
        if np.size(rr) == np.size(row):
            rr = row
        if np.size(xx) * np.size(rr) != np.size(x):
            return False, x, row, args
        outargs = [arg[0:np.size(xx)] for arg in args]
        return True, xx, rr, outargs

    def _rvs(self, *args, size=None, random_state=None):
        # Use basic inverse cdf algorithm for RV generation as default.
        U = random_state.uniform(size=size)
        Y = self._ppf(U, *args)
        if size is None:  #pragma: no cover
            return Y
        return Y.reshape(size)

    def _argcheck(self, *args):
        """Default check for correct values on args and keywords.
        Returns condition array of 1's where arguments are correct and
         0's where they are not.
        """
        cond = 1
        if args:
            cond = np.logical_and(cond, np.logical_and(asarray(args[0]) >= 0, asarray(args[0]) < self._npdf))
        return np.atleast_1d(cond)

    def freeze(self, *args, **kwds):
        """Freeze the distribution for the given arguments.

        Parameters
        ----------
        arg1, arg2, arg3,... : array_like
            The shape parameter(s) for the distribution.  Should include all
            the non-optional arguments, may include ``loc`` and ``scale``.

        Returns
        -------
        rv_frozen : rv_frozen instance
            The frozen distribution.
        """
        return rv_frozen_rows(self, self._shape, *args, **kwds)

    @classmethod
    def create_gen(cls, **kwds):
        """Create and return a `scipy.stats.rv_continuous` object using the
        keyword arguemntets provided"""
        return (cls(**kwds), {})

    def _scipy_version_warning(self):
        import scipy  #pylint: disable=import-outside-toplevel
        scipy_version = scipy.__version__
        vtuple = scipy_version.split('.')
        if int(vtuple[0]) > 1 or int(vtuple[1]) > 7:
            return
        raise DeprecationWarning(f"Ensemble.moments will not work correctly with scipy version < 1.8.0, you have {scipy_version}")  #pragma: no cover

    def moment(self, n, *args, **kwds):
        """Returns the moments request moments for all the PDFs.

        This used to call a hacked version `Pdf_gen._moment_fix` which can handle cases of multiple PDFs.
        Now it prints a deprication warning for scipy < 1.8

        Parameters
        ----------
        n : int
            Order of the moment

        Returns
        -------
        moments : array_like
            The requested moments
        """
        self._scipy_version_warning()
        return rv_continuous.moment(self, n, *args, **kwds)



class Pdf_gen_wrap(Pdf_gen):
    """Mixin class to extend `scipy.stats.rv_continuous` with
    information needed for `qp` for analytic distributions.

    """
    def __init__(self, *args, **kwargs):
        """C'tor"""
        # pylint: disable=no-member,protected-access
        super().__init__(*args, **kwargs)
        self._other_init(*args, **kwargs)


    def _my_freeze(self, *args, **kwds):
        """Freeze the distribution for the given arguments.

        Parameters
        ----------
        arg1, arg2, arg3,... : array_like
            The shape parameter(s) for the distribution.  Should include all
            the non-optional arguments, may include ``loc`` and ``scale``.

        Returns
        -------
        rv_frozen : rv_frozen instance
            The frozen distribution.
        """
        # pylint: disable=no-member,protected-access
        args, loc, scale = self._parse_args(*args, **kwds)
        x, loc, scale = map(asarray, (1, loc, scale))
        x = np.asarray((x - loc)/scale)
        args = tuple(map(asarray, args))
        cond0 = np.atleast_1d(self._argcheck(*args)) & (scale > 0)
        cond1 = self._support_mask(x, *args) & (scale > 0)
        cond = cond0 & cond1
        return rv_frozen_func(self, cond.shape[:-1], *args, **kwds)

    def _my_argcheck(self, *args):
        # pylint: disable=no-member,protected-access
        return np.atleast_1d(self._other_argcheck(*args))


    @classmethod
    def get_allocation_kwds(cls, npdf, **kwargs):
        return {key:((npdf,1), val.dtype) for key, val in kwargs.items()}


    @classmethod
    def add_mappings(cls):
        """
        Add this classes mappings to the conversion dictionary
        """
        cls._add_creation_method(cls.create, None)



"""Utility functions for the qp package"""

import sys
import numpy as np

from scipy import stats as sps
from scipy.interpolate import interp1d

epsilon = sys.float_info.epsilon
infty = sys.float_info.max * epsilon
lims = (epsilon, 1.)

CASE_PRODUCT = 0
CASE_FACTOR = 1
CASE_2D = 2
CASE_FLAT = 3

def safelog(arr, threshold=epsilon):
    """
    Takes the natural logarithm of an array of potentially non-positive numbers

    Parameters
    ----------
    arr: numpy.ndarray, float
        values to be logged
    threshold: float
        small, positive value to replace zeros and negative numbers

    Returns
    -------
    logged: numpy.ndarray
        logarithms, with approximation in place of zeros and negative numbers
    """
    return np.log(np.array(arr).clip(threshold, np.inf))


_ = """
def normalize_quantiles(in_data, threshold=epsilon, vb=False):
    Evaluates PDF from quantiles including endpoints from linear extrapolation

    Parameters
    ----------
    in_data: tuple, numpy.ndarray, float
        tuple of CDF values iy corresponding to quantiles and the points x at
        which those CDF values are achieved
    threshold: float, optional
        optional minimum threshold for PDF
    vb: boolean, optional
        be careful and print progress to stdout?

    Returns
    -------
    out_data: tuple, ndarray, float
        tuple of values x at which CDF is achieved, including extrema, and
        normalized PDF values y at x
    (iy, x) = in_data
    (xs, ys) = evaluate_quantiles((iy, x), vb=vb)
    # xs = xs[1:-1]
    # ys = ys[1:-1]
    x_min = xs[0] - 2 * iy[0] / ys[0]
    x_max = xs[-1] + 2 * (1. - iy[-1]) / ys[-1]
    xs = sandwich(xs, (x_min, x_max))
    ys = sandwich(ys, (threshold, threshold))
    out_data = (xs, ys)
    return out_data
"""

def edge_to_center(edges):
    """Return the centers of a set of bins given the edges"""
    return 0.5*(edges[1:] + edges[:-1])


def bin_widths(edges):
    """Return the widths of a set of bins given the edges"""
    return edges[1:] - edges[:-1]


def get_bin_indices(bins, x):
    """Return the bin indexes for a set of values

    If the bins are equal width this will use arithmatic,
    If the bins are not equal width this will use a binary search
    """
    widths = bin_widths(bins)
    n_bins = np.size(bins) - 1
    if np.allclose(widths, widths[0]):
        idx = np.atleast_1d(np.floor((x-bins[0])/widths[0]).astype(int))
    else:
        idx = np.atleast_1d(np.searchsorted(bins, x, side='left')-1)
    mask = (idx >= 0) * (idx < bins.size-1)
    np.putmask(idx, 1-mask, 0)
    xshape = np.shape(x)
    return idx.reshape(xshape).clip(0, n_bins-1), mask.reshape(xshape)


def normalize_interp1d(xvals, yvals):
    """
    Normalize a set of 1D interpolators

    Parameters
    ----------
    xvals : array-like
        X-values used for the interpolation
    yvals : array-like
        Y-values used for the interpolation

    Returns
    -------
    ynorm: array-like
        Normalized y-vals
    """
    #def row_integral(irow):
    #    return quad(interp1d(xvals[irow], yvals[irow], **kwargs), limits[0], limits[1])[0]

    #vv = np.vectorize(row_integral)
    #integrals = vv(np.arange(xvals.shape[0]))
    integrals = np.sum(xvals[:,1:]*yvals[:,1:] - xvals[:,:-1]*yvals[:,1:], axis=1)
    return (yvals.T / integrals).T


def build_kdes(samples, **kwargs):
    """
    Build a set of Gaussian Kernal Density Estimates

    Parameters
    ----------
    samples : array-like
        X-values used for the spline

    Keywords
    --------
    Passed to the `scipy.stats.gaussian_kde` constructor

    Returns
    -------
    kdes : list of `scipy.stats.gaussian_kde` objects
    """
    return [ sps.gaussian_kde(row, **kwargs) for row in samples ]



def evaluate_kdes(xvals, kdes):
    """
    Build a evaluate a set of kdes

    Parameters
    ----------
    xvals : array_like
        X-values used for the spline
    kdes : list of `sps.gaussian_kde`
        The kernel density estimates

    Returns
    -------
    yvals : array_like
        The kdes evaluated at the xvamls
    """
    return np.vstack([kde(xvals) for kde in kdes])


def get_eval_case(x, row):
    """ Figure out which of the various input formats scipy.stats has passed us

    Parameters
    ----------
    x : array_like
        Pdf x-vals
    row : array_like
        Pdf row indices

    Returns
    -------
    case : `int`
        The case code
    xx : array_like
        The x-values properly shaped
    rr : array_like
        The y-values, properly shaped

    Notes
    -----
    The cases are:

    CASE_FLAT : x, row have shapes (n), (n) and do not factor
    CASE_FACTOR : x, row have shapes (n), (n) but can be factored to shapes (1, nx) and (npdf, 1) (i.e., they were flattend by scipy) 
    CASE_PRODUCT : x, row have shapes (1, nx) and (npdf, 1)
    CASE_2D : x, row have shapes (npdf, nx) and (npdf, nx)
    """
    nd_x = np.ndim(x)
    nd_row = np.ndim(row)
    #if nd_x > 2 or nd_row > 2:  #pragma: no cover
    #    raise ValueError("Too many dimensions: x(%s), row(%s)" % (np.shape(x), np.shape(row)))
    if nd_x >= 2 and nd_row != 1:
        return CASE_2D, x, row
    if nd_x >= 2 and nd_row == 1:  #pragma: no cover
        raise ValueError("Dimension mismatch: x(%s), row(%s)" % (np.shape(x), np.shape(row)))
    if nd_row >= 2:
        return CASE_PRODUCT, x, row
    if np.size(x) == 1 or np.size(row) == 1:
        return CASE_FLAT, x, row
    xx = np.unique(x)
    rr = np.unique(row)
    if np.size(xx) == np.size(x):
        xx = x
    if np.size(rr) == np.size(row):
        rr = row
    if np.size(xx) * np.size(rr) != np.size(x):
        return CASE_FLAT, x, row
    return CASE_FACTOR, xx, np.expand_dims(rr, -1)


def evaluate_hist_x_multi_y_flat(x, row, bins, vals, derivs=None):  #pragma: no cover
    """
    Evaluate a set of values from histograms

    Parameters
    ----------
    x : array_like (n)
        X values to interpolate at
    row : array_like (n)
        Which rows to interpolate at
    bins : array_like (N+1)
        'x' bin edges
    vals : array_like (npdf, N)
        'y' bin contents

    Returns
    -------
    out : array_like (n)
        The histogram values
    """
    assert np.ndim(x) < 2 and np.ndim(row) < 2
    idx, mask = get_bin_indices(bins, x)
    if derivs is None:
        deltas = np.zeros(idx.shape)
    else:
        deltas = x - bins[idx]
    def evaluate_row(idxv, maskv, rv, delta):
        if derivs is None:
            return np.where(maskv, vals[rv, idxv], 0)
        return np.where(maskv, vals[rv, idxv] + delta*derivs[rv, idxv], 0)
    vv = np.vectorize(evaluate_row)
    return vv(idx, mask, row, deltas)


def evaluate_hist_x_multi_y_product(x, row, bins, vals, derivs=None):  #pragma: no cover
    """
    Evaluate a set of values from histograms

    Parameters
    ----------
    x : array_like (npts)
        X values to interpolate at
    row : array_like (npdf, 1)
        Which rows to interpolate at
    bins : array_like (N+1)
        'x' bin edges
    vals : array_like (npdf, N)
        'y' bin contents

    Returns
    -------
    out : array_like (npdf, npts)
        The histogram values
    """
    #assert np.ndim(x) < 2 and np.ndim(row) == 2
    idx, mask0 = get_bin_indices(bins, x)
    mask = np.ones(row.shape) * mask0
    if derivs is None:
        return np.where(mask, vals[:,idx][np.squeeze(row)], 0)
    deltas = x - bins[idx]
    return np.where(mask, vals[:,idx][np.squeeze(row)] + deltas*derivs[:,idx][np.squeeze(row)] , 0)


def evaluate_hist_x_multi_y_2d(x, row, bins, vals, derivs=None):  #pragma: no cover
    """
    Evaluate a set of values from histograms

    Parameters
    ----------
    x : array_like (npdf, npts)
        X values to interpolate at
    row : array_like (npdf, 1)
        Which rows to interpolate at
    bins : array_like (N+1)
        'x' bin edges
    vals : array_like (npdf, N)
        'y' bin contents

    Returns
    -------
    out : array_like (npdf, npts)
        The histogram values
    """
    assert np.ndim(x) >= 2 and np.ndim(row) >= 2
    idx, mask = get_bin_indices(bins, x)
    if derivs is None:
        deltas = np.zeros(idx.shape)
    else:
        deltas = x - bins[idx]

    def evaluate_row(idxv, maskv, rv, delta):
        if derivs is None:
            return np.where(maskv, vals[rv, idxv], 0)
        return np.where(maskv, vals[rv, idxv] + delta*derivs[rv, idxv], 0)
    vv = np.vectorize(evaluate_row)
    return vv(idx, mask, row, deltas)


def evaluate_hist_x_multi_y(x, row, bins, vals, derivs=None):
    """
    Evaluate a set of values from histograms

    Parameters
    ----------
    x : array_like
        X values to interpolate at
    row : array_like
        Which rows to interpolate at
    bins : array_like (N+1)
        'x' bin edges
    vals : array_like (npdf, N)
        'y' bin contents

    Returns
    -------
    out : array_like
        The histogram values

    Notes
    -----
    Depending on the shape of 'x' and 'row' this will
    use one of the three specific implementations.
    """
    case_idx, xx, rr = get_eval_case(x, row)
    if case_idx in [CASE_PRODUCT, CASE_FACTOR]:
        return evaluate_hist_x_multi_y_product(xx, rr, bins, vals, derivs)
    if case_idx == CASE_2D:
        return evaluate_hist_x_multi_y_2d(xx, rr, bins, vals, derivs)
    return evaluate_hist_x_multi_y_flat(xx, rr, bins, vals, derivs)


def evaluate_hist_multi_x_multi_y_flat(x, row, bins, vals, derivs=None):  #pragma: no cover
    """
    Evaluate a set of values from histograms

    Parameters
    ----------
    x : array_like (n)
        X values to interpolate at
    row : array_like (n)
        Which rows to interpolate at
    bins : array_like (npdf, N+1)
        'x' bin edges
    vals : array_like (npdf, N)
        'y' bin contents

    Returns
    -------
    out : array_like (n)
        The histogram values
    """
    def evaluate_row(xv, rv):
        bins_row = bins[rv]
        idx, mask = get_bin_indices(bins_row, xv)
        delta = xv - bins_row[idx]
        if derivs is None:
            return np.where(mask, vals[rv, idx], 0)
        return np.where(mask, vals[rv, idx] + delta*derivs[rv, idx], 0)
    vv = np.vectorize(evaluate_row)
    return vv(x, row)


def evaluate_hist_multi_x_multi_y_product(x, row, bins, vals, derivs=None):  #pragma: no cover
    """
    Evaluate a set of values from histograms

    Parameters
    ----------
    x : array_like (npts)
        X values to interpolate at
    row : array_like (npdf, 1)
        Which rows to interpolate at
    bins : array_like (npdf, N+1)
        'x' bin edges
    vals : array_like (npdf, N)
        'y' bin contents

    Returns
    -------
    out : array_like (npdf, npts)
        The histogram values
    """
    def evaluate_row(rv):
        bins_flat = bins[rv].flatten()
        idx, mask = get_bin_indices(bins_flat, x)
        delta = x - bins_flat[idx]
        if derivs is None:
            return np.where(mask, np.squeeze(vals[rv])[idx], 0).flatten()
        return np.where(mask, np.squeeze(vals[rv])[idx] + delta* np.squeeze(derivs[rv])[idx], 0)
    vv = np.vectorize(evaluate_row, signature="(1)->(%i)" % (x.size))
    return vv(row)


def evaluate_hist_multi_x_multi_y_2d(x, row, bins, vals, derivs=None):  #pragma: no cover
    """
    Evaluate a set of values from histograms

    Parameters
    ----------
    x : array_like (npdf, npts)
        X values to interpolate at
    row : array_like (npdf, 1)
        Which rows to interpolate at
    bins : array_like (npdf, N+1)
        'x' bin edges
    vals : array_like (npdf, N)
        'y' bin contents

    Returns
    -------
    out : array_like (npdf, npts)
        The histogram values
    """
    nx = np.shape(x)[-1]
    def evaluate_row(rv, xv):
        flat_bins = bins[rv].flatten()
        idx, mask = get_bin_indices(flat_bins, xv)
        delta = xv - flat_bins[idx]
        if derivs is None:
            return np.where(mask, np.squeeze(vals[rv])[idx], 0).flatten()
        return np.where(mask, np.squeeze(vals[rv])[idx] + delta*np.squeeze(derivs[rv])[idx], 0).flatten()
    vv = np.vectorize(evaluate_row, signature="(1),(%i)->(%i)" % (nx, nx))
    return vv(row, x)

def evaluate_hist_multi_x_multi_y(x, row, bins, vals, derivs=None):
    """
    Evaluate a set of values from histograms

    Parameters
    ----------
    x : array_like
        X values to interpolate at
    row : array_like
        Which rows to interpolate at
    bins : array_like (npdf, N+1)
        'x' bin edges
    vals : array_like (npdf, N)
        'y' bin contents

    Returns
    -------
    out : array_like
        The histogram values
    """
    case_idx, xx, rr = get_eval_case(x, row)
    if case_idx in [CASE_PRODUCT, CASE_FACTOR]:
        return evaluate_hist_multi_x_multi_y_product(xx, rr, bins, vals, derivs)
    if case_idx == CASE_2D:
        return evaluate_hist_multi_x_multi_y_2d(xx, rr, bins, vals, derivs)
    return evaluate_hist_multi_x_multi_y_flat(xx, rr, bins, vals, derivs)


def interpolate_x_multi_y_flat(x, row, xvals, yvals, **kwargs):
    """
    Interpolate a set of values

    Parameters
    ----------
    x : array_like (n)
        X values to interpolate at
    row : array_like (n)
        Which rows to interpolate at
    xvals : array_like (npts)
        X-values used for the interpolation
    yvals : array_like (npdf, npts)
        Y-avlues used for the inteolation

    Returns
    -------
    vals : array_like (npdf, n)
        The interpoalted values
    """
    def single_row(xv, rv):
        return interp1d(xvals, yvals[rv], **kwargs)(xv)
    vv = np.vectorize(single_row)
    return vv(x, row)


def interpolate_x_multi_y_product(x, row, xvals, yvals, **kwargs):
    """
    Interpolate a set of values

    Parameters
    ----------
    x : array_like (n)
        X values to interpolate at
    row : array_like (npdf, 1)
        Which rows to interpolate at
    xvals : array_like (npts)
        X-values used for the interpolation
    yvals : array_like (npdf, npts)
        Y-avlues used for the inteolation

    Returns
    -------
    vals : array_like (npdf, n)
        The interpoalted values
    """
    rr = np.squeeze(row)
    return interp1d(xvals, yvals[rr], **kwargs)(x)


def interpolate_x_multi_y_2d(x, row, xvals, yvals, **kwargs):
    """
    Interpolate a set of values

    Parameters
    ----------
    x : array_like (npdf, n)
        X values to interpolate at
    row : array_like (npdf, 1)
        Which rows to interpolate at
    xvals : array_like (npts)
        X-values used for the interpolation
    yvals : array_like (npdf, npts)
        Y-avlues used for the inteolation

    Returns
    -------
    vals : array_like (npdf, n)
        The interpoalted values
    """
    nx = np.shape(x)[-1]
    def evaluate_row(rv, xv):
        return interp1d(xvals, yvals[rv], **kwargs)(xv)
    vv = np.vectorize(evaluate_row, signature="(1),(%i)->(%i)" % (nx, nx))
    return vv(row, x)


def interpolate_x_multi_y(x, row, xvals, yvals, **kwargs):
    """
    Interpolate a set of values

    Parameters
    ----------
    x : array_like (npdf, n)
        X values to interpolate at
    row : array_like (npdf, 1)
        Which rows to interpolate at
    xvals : array_like (npts)
        X-values used for the interpolation
    yvals : array_like (npdf, npts)
        Y-avlues used for the inteolation

    Returns
    -------
    vals : array_like
        The interpoalted values
    """
    case_idx, xx, rr = get_eval_case(x, row)
    if case_idx in [CASE_PRODUCT, CASE_FACTOR]:
        return interpolate_x_multi_y_product(xx, rr, xvals, yvals, **kwargs)
    if case_idx == CASE_2D:
        return interpolate_x_multi_y_2d(xx, rr, xvals, yvals, **kwargs)
    return interpolate_x_multi_y_flat(xx, rr, xvals, yvals, **kwargs)


def interpolate_multi_x_multi_y_flat(x, row, xvals, yvals, **kwargs):
    """
    Interpolate a set of values

    Parameters
    ----------
    x : array_like (n)
        X values to interpolate at
    row : array_like (n)
        Which rows to interpolate at
    xvals : array_like (npdf, npts)
        X-values used for the interpolation
    yvals : array_like (npdf, npts)
        Y-avlues used for the inteolation

    Returns
    -------
    vals : array_like (npdf, n)
        The interpoalted values
    """
    def single_row(xv, rv):
        return interp1d(xvals[rv], yvals[rv], **kwargs)(xv)
    vv = np.vectorize(single_row)
    return vv(x, row)


def interpolate_multi_x_multi_y_product(x, row, xvals, yvals, **kwargs):
    """
    Interpolate a set of values

    Parameters
    ----------
    x : array_like (n)
        X values to interpolate at
    row : array_like (npdf, 1)
        Which rows to interpolate at
    xvals : array_like (npdf, npts)
        X-values used for the interpolation
    yvals : array_like (npdf, npts)
        Y-avlues used for the inteolation

    Returns
    -------
    vals : array_like (npdf, n)
        The interpoalted values
    """
    rr = np.squeeze(row)
    nx = np.shape(x)[-1]
    def single_row(rv):
        return interp1d(xvals[rv], yvals[rv], **kwargs)(x)
    vv = np.vectorize(single_row, signature="()->(%i)" % (nx))
    return vv(rr)


def interpolate_multi_x_multi_y_2d(x, row, xvals, yvals, **kwargs):
    """
    Interpolate a set of values

    Parameters
    ----------
    x : array_like (npdf, n)
        X values to interpolate at
    row : array_like (npdf, 1)
        Which rows to interpolate at
    xvals : array_like (npdf, npts)
        X-values used for the interpolation
    yvals : array_like (npdf, npts)
        Y-avlues used for the inteolation

    Returns
    -------
    vals : array_like (npdf, n)
        The interpoalted values
    """
    nx = np.shape(x)[-1]
    def evaluate_row(rv, xv):
        return interp1d(xvals[rv], yvals[rv], **kwargs)(xv)
    vv = np.vectorize(evaluate_row, signature="(),(%i)->(%i)" % (nx, nx))
    return vv(np.squeeze(row), x)


def interpolate_multi_x_multi_y(x, row, xvals, yvals, **kwargs):
    """
    Interpolate a set of values

    Parameters
    ----------
    x : array_like (npdf, n)
        X values to interpolate at
    row : array_like (npdf, 1)
        Which rows to interpolate at
    xvals : array_like (npdf, npts)
        X-values used for the interpolation
    yvals : array_like (npdf, npts)
        Y-avlues used for the inteolation

    Returns
    -------
    vals : array_like
        The interpoalted values
    """
    case_idx, xx, rr = get_eval_case(x, row)
    if case_idx in [CASE_PRODUCT, CASE_FACTOR]:
        return interpolate_multi_x_multi_y_product(xx, rr, xvals, yvals, **kwargs)
    if case_idx == CASE_2D:
        return interpolate_multi_x_multi_y_2d(xx, rr, xvals, yvals, **kwargs)
    return interpolate_multi_x_multi_y_flat(xx, rr, xvals, yvals, **kwargs)


def interpolate_multi_x_y_flat(x, row, xvals, yvals, **kwargs):
    """
    Interpolate a set of values

    Parameters
    ----------
    x : array_like (n)
        X values to interpolate at
    row : array_like (n)
        Which rows to interpolate at
    xvals : array_like (npdf, npts)
        X-values used for the interpolation
    yvals : array_like (npdf)
        Y-avlues used for the inteolation

    Returns
    -------
    vals : array_like (npdf, n)
        The interpoalted values
    """
    def single_row(xv, rv):
        return interp1d(xvals[rv], yvals, **kwargs)(xv)
    vv = np.vectorize(single_row)
    return vv(x, row)


def interpolate_multi_x_y_product(x, row, xvals, yvals, **kwargs):
    """
    Interpolate a set of values

    Parameters
    ----------
    x : array_like (n)
        X values to interpolate at
    row : array_like (npdf, 1)
        Which rows to interpolate at
    xvals : array_like (npdf, npts)
        X-values used for the interpolation
    yvals : array_like (npdf)
        Y-avlues used for the inteolation

    Returns
    -------
    vals : array_like (npdf, n)
        The interpoalted values
    """
    rr = np.squeeze(row)
    nx = np.shape(x)[-1]
    def single_row(rv):
        return interp1d(xvals[rv], yvals, **kwargs)(x)
    vv = np.vectorize(single_row, signature="()->(%i)" % (nx))
    return vv(rr)


def interpolate_multi_x_y_2d(x, row, xvals, yvals, **kwargs):
    """
    Interpolate a set of values

    Parameters
    ----------
    x : array_like (npdf, n)
        X values to interpolate at
    row : array_like (npdf, 1)
        Which rows to interpolate at
    xvals : array_like (npdf, npts)
        X-values used for the interpolation
    yvals : array_like (npdf)
        Y-avlues used for the inteolation

    Returns
    -------
    vals : array_like (npdf, n)
        The interpoalted values
    """
    nx = np.shape(x)[-1]
    def evaluate_row(rv, xv):
        return interp1d(xvals[rv], yvals, **kwargs)(xv)
    vv = np.vectorize(evaluate_row, signature="(),(%i)->(%i)" % (nx, nx))
    return vv(np.squeeze(row), x)


def interpolate_multi_x_y(x, row, xvals, yvals, **kwargs):
    """
    Interpolate a set of values

    Parameters
    ----------
    x : array_like (npdf, n)
        X values to interpolate at
    row : array_like (npdf, 1)
        Which rows to interpolate at
    xvals : array_like (npdf, npts)
        X-values used for the interpolation
    yvals : array_like (npdf)
        Y-avlues used for the inteolation

    Returns
    -------
    vals : array_like
        The interpoalted values
    """
    case_idx, xx, rr = get_eval_case(x, row)
    if case_idx in [CASE_PRODUCT, CASE_FACTOR]:
        return interpolate_multi_x_y_product(xx, rr, xvals, yvals, **kwargs)
    if case_idx == CASE_2D:
        return interpolate_multi_x_y_2d(xx, rr, xvals, yvals, **kwargs)
    return interpolate_multi_x_y_flat(xx, rr, xvals, yvals, **kwargs)


def profile(x_data, y_data, x_bins, std=True):
    """Make a 'profile' plot

    Paramters
    ---------
    x_data : array_like (n)
        The x-values
    y_data : array_like (n)
        The y-values
    x_bins : array_like (nbins+1)
        The values of the bin edges
    std : bool
        If true, return the standard deviations, if false return the errors on the means

    Returns
    -------
    vals : array_like (nbins)
        The means
    errs : array_like (nbins)
        The standard deviations or errors on the means
    """
    idx, mask = get_bin_indices(x_bins, x_data)
    count = np.zeros(x_bins.size-1)
    vals = np.zeros(x_bins.size-1)
    errs = np.zeros(x_bins.size-1)
    for i in range(x_bins.size-1):
        mask_col = mask * (idx == i)
        count[i] = mask_col.sum()
        if mask_col.sum() == 0:  #pragma: no cover
            vals[i] = np.nan
            errs[i] = np.nan
            continue
        masked_vals = y_data[mask_col]
        vals[i] = masked_vals.mean()
        errs[i] = masked_vals.std()
    if not std:
        errs /= np.sqrt(count)
    return vals, errs

def reshape_to_pdf_size(vals, split_dim):
    """Reshape an array to match the number of PDFs in a distribution
    Parameters
    ----------
    vals : array
        The input array
    split_dim : int
        The dimension at which to split between pdf indices and per_pdf indices
    Returns
    -------
    out : array
        The reshaped array
    """
    in_shape = np.shape(vals)
    npdf = np.product(in_shape[:split_dim]).astype(int)
    per_pdf = in_shape[split_dim:]
    out_shape = np.hstack([npdf, per_pdf])
    return vals.reshape(out_shape)


def reshape_to_pdf_shape(vals, pdf_shape, per_pdf):
    """Reshape an array to match the shape of PDFs in a distribution

    Parameters
    ----------
    vals : array
        The input array
    pdf_shape : int
        The shape for the pdfs
    per_pdf : int or array_like
        The shape per pdf

    Returns
    -------
    out : array
        The reshaped array
    """
    outshape = np.hstack([pdf_shape, per_pdf])
    return vals.reshape(outshape)



"""This module contains functions copied from the 1.10.0dev branch of Scipy.
The original code can be found here: https://github.com/scipy/scipy/blob/maintenance/1.10.x/scipy/stats/_fit.py#L722
The original Scipy 1.10.0dev code is wrapped with a function that prevents passing more than 1 distribution at a time.
So, temporarily, we'll make use of the underlying, vectorized functions.
Fortunately, the vectorized Scipy code works without modification. 
Once Scipy 1.10 is made available, we can swap out the copied functions for those in Scipy 1.10.
"""

import numpy as np

def _anderson_darling(dist, data):
    x = np.sort(data, axis=-1)
    n = data.shape[-1]
    i = np.arange(1, n+1)
    Si = (2*i - 1)/n * (dist.logcdf(x) + dist.logsf(x[..., ::-1]))
    S = np.sum(Si, axis=-1)
    return -n - S

def _kolmogorov_smirnov(dist, data):
    x = np.sort(data, axis=-1)
    cdfvals = dist.cdf(x)
    Dplus = _compute_dplus(cdfvals)  # always works along last axis
    Dminus = _compute_dminus(cdfvals)
    return np.maximum(Dplus, Dminus)

def _compute_dplus(cdfvals):  # adapted from _stats_py before gh-17062
    n = cdfvals.shape[-1]
    return (np.arange(1.0, n + 1) / n - cdfvals).max(axis=-1)

def _compute_dminus(cdfvals, axis=-1):
    n = cdfvals.shape[-1]
    return (cdfvals - np.arange(0.0, n)/n).max(axis=axis)

def _cramer_von_mises(dist, data):
    x = np.sort(data, axis=-1)
    n = data.shape[-1]
    cdfvals = dist.cdf(x)
    u = (2*np.arange(1, n+1) - 1)/(2*n)
    w = 1 / (12*n) + np.sum((u - cdfvals)**2, axis=-1)
    return w

# The following methods can be replaced by:
# scipy.stats._fit._anderson_darling,
# scipy.stats._fit._cramer_von_mises, and
# scipy.stats._fit._kolmogorov_smirnov when Scipy 1.10 is available.
goodness_of_fit_metrics = {
    'ad': _anderson_darling,
    'cvm': _cramer_von_mises,
    'ks': _kolmogorov_smirnov
}



"""qp is a library for managing and converting between different representations of distributions"""

import os
from .version import __version__

from .spline_pdf import *
from .hist_pdf import *
from .interp_pdf import *
from .quant_pdf import *
from .mixmod_pdf import *
from .sparse_pdf import *
from .scipy_pdfs import *
from .packed_interp_pdf import *
from .ensemble import Ensemble
from .factory import instance, add_class, create, read, convert, concatenate, iterator, data_length, from_tables

from . import utils

from . import packing_utils

from . import test_funcs



"""
Unit tests for PDF class
"""
import copy
import numpy as np
import unittest
import qp
from qp import test_data
import os
import sys
import pytest

from qp.test_funcs import assert_all_small, assert_all_close, build_ensemble
from qp.plotting import init_matplotlib
from mpi4py import MPI
import h5py


@pytest.mark.skipif(not h5py.get_config().mpi, reason="Do not have parallel version of hdf5py")
class EnsembleTestCase(unittest.TestCase):
    """ Class to test qp.Ensemble functionality """

    def setUp(self):
        """
        Make any objects that are used in multiple tests.
        """

    def tearDown(self):
        "Clean up any mock data files created by the tests."

    @staticmethod
    def _run_ensemble_funcs(ens_type, ens, xpts):
        """Run the test for a practicular class"""
        comm = MPI.COMM_WORLD
        mpi_rank = comm.Get_rank()
        mpi_size = comm.Get_size()
        zmode = ens.mode(grid=np.linspace(-3,3,100))
        ens.set_ancil(dict(zmode=zmode,ones=np.ones(ens.npdf)))
        group, fout = ens.initializeHdf5Write(f"testwrite_{ens_type}.hdf5", ens.npdf*mpi_size, comm)
        ens.writeHdf5Chunk(group, mpi_rank*ens.npdf, (mpi_rank+1)*ens.npdf)
        ens.finalizeHdf5Write(fout)

        readens = qp.read(f"testwrite_{ens_type}.hdf5")
        assert sum(readens.ancil['ones']) == mpi_size*ens.npdf
        assert len(readens.ancil['zmode']) == mpi_size*ens.npdf
        assert readens.metadata().keys() == ens.metadata().keys()
        assert readens.objdata().keys() == ens.objdata().keys()

        test_grid = grid=np.linspace(-3,3,100)        
        itr = qp.iterator(f"testwrite_{ens_type}.hdf5", 10, mpi_rank, mpi_size)
        for start, end, ens_i in itr:
            assert np.allclose(readens[start:end].pdf(test_grid), ens_i.pdf(test_grid))
        
        if mpi_rank == 0:
            os.remove(f"testwrite_{ens_type}.hdf5")

    def test_parallel_norm(self):
        """ Run the ensemble tests on an ensemble of scipy.stats.norm distributions """
        key = 'norm'
        cls_test_data = qp.stats.norm_gen.test_data[key]  #pylint: disable=no-member
        ens_norm = build_ensemble(cls_test_data)
        self._run_ensemble_funcs('norm', ens_norm, cls_test_data['test_xvals'])

    def test_parallel_hist(self):
        """ Run the ensemble tests on an ensemble of qp.hist distributions """
        key = 'hist'
        qp.hist_gen.make_test_data()
        cls_test_data = qp.hist_gen.test_data[key]
        ens_h = build_ensemble(cls_test_data)
        self._run_ensemble_funcs('hist', ens_h, cls_test_data['test_xvals'])

    def test_parallel_interp(self):
        """ Run the ensemble tests on an ensemble of qp.interp distributions """
        key = 'interp'
        qp.interp_gen.make_test_data()
        cls_test_data = qp.interp_gen.test_data[key]
        ens_i = build_ensemble(cls_test_data)
        self._run_ensemble_funcs('interp', ens_i, cls_test_data['test_xvals'])


if __name__ == '__main__':
    unittest.main()



from typing import List

import numpy as np
from scipy.interpolate import InterpolatedUnivariateSpline

from qp.quantile_pdf_constructors.abstract_pdf_constructor import AbstractQuantilePdfConstructor

class CdfSplineDerivative(AbstractQuantilePdfConstructor):
    """Implements an interpolation algorithm based on a list of quantiles and locations.
    First we fit a spline to the (quantile,location) pairs. Then evaluate the derivative
    of the spline. This represents a reconstruction of the original PDF from which the
    quantiles and locations were selected.

    Calling `cdf_spline.interpolate(grid)` will evaluate the spline derivatives at the
    provided grid values.
    """
    def __init__(self, quantiles:List[float], locations: List[List[float]]) -> None:
        """Constructor to instantiate this class.

        Parameters
        ----------
        quantiles : List[float]
            List of n quantile values in the range (0,1).
        locations : List[List[float]]
            List of m Lists, each containing n values corresponding to the
            y-value of the PPF function at the same quantile index.
        """
        self._quantiles = quantiles
        self._locations = np.atleast_2d(locations)

        # A list of interpolation functions (spline derivatives fit to quant,loc pairs)
        self._interpolation_functions = None

    def prepare_constructor(self, spline_order:int = 3) -> None:
        """Calculate the fit spline derivative for each of the original distributions
        This function is the least performant - for reference, on a M1 Mac,
        it requires about 30 seconds to produce an output given
        shape(locations) = (1_000_000, 30).

        Note: we are aware that the edges of the resulting pdf are showing an
        elephant foot.

        Parameters
        ----------
        spline_order : int
            Defines the order of the spline fit, defaults to 4
        """
        number_of_locations = len(self._locations[:,0])

        # ! create an issue (or fix) if the spline fit fails, can fall back to a simpler interpolator ???
        self._interpolation_functions = [
            InterpolatedUnivariateSpline(self._locations[i,:], self._quantiles, k=spline_order, ext=1).derivative()
            for i in range(0,number_of_locations)
        ]

    def construct_pdf(self, grid: List[float], row: List[int] = None) -> List[List[float]]:
        """Evaluate the fit spline derivative at each of the grid values

        Parameters
        ----------
        grid : List[float]
            The x values to pass to self._interpolation_functions
        row : List[int], optional
            Defines which interpolation_functions to return values for, by default None

        Returns
        -------
        List[List[float]]
            The lists of y values returned from self._interpolation_functions
        """

        # Generate the fitted spline derivatives if they don't already exist.
        if self._interpolation_functions is None:
            self.prepare_constructor()

        # Support the use of `row` as a filter. If row is None, do nothing,
        # otherwise, return a subset of the rows.
        selected_interpolation_functions = self._interpolation_functions
        if row is not None:
            selected_interpolation_functions = map(self._interpolation_functions.__getitem__, np.unique(row))

        # For each of the fit spline derivative, calculate y value given the grid (or x) values.
        # Note: This implementation uses list comprehension, there might be a faster way.
        # Using np.vectorize was attempted, however:
        # 1) np.vectorize doesn't work well for a list of functions (i.e. self._interpolation_functions)
        # 2) np.vectorize doesn't provide any any performance improvements over list comprehension
        return np.asarray([func(grid) for func in selected_interpolation_functions])

    def debug(self):
        """This is a debugging utility that is meant to return intermediate calculation values
        to make it easier to visualize and debug the reconstruction algorithm.

        Returns
        -------
            _quantiles :
                Input during constructor instantiation
            _locations :
                Input during constructor instantiation
            _interpolation_functions :
                The list of analytic derivatives of splines fit to the input data
        """
        return self._quantiles, self._locations, self._interpolation_functions



"""This module implements a PDT distribution sub-class using a Gaussian mixture model
"""

import numpy as np

from scipy.stats import rv_continuous
from scipy import stats as sps


from qp.pdf_gen import Pdf_rows_gen
from qp.conversion_funcs import extract_mixmod_fit_samples
from qp.test_data import WEIGHT_MIXMOD, MEAN_MIXMOD, STD_MIXMOD, TEST_XVALS
from qp.factory import add_class
from qp.utils import reshape_to_pdf_size, interpolate_multi_x_y, get_eval_case


class mixmod_gen(Pdf_rows_gen):
    """Mixture model based distribution

    Notes
    -----
    This implements a PDF using a Gaussian Mixture model

    The relevant data members are:

    means:  (npdf, ncomp) means of the Gaussians
    stds:  (npdf, ncomp) standard deviations of the Gaussians
    weights: (npdf, ncomp) weights for the Gaussians

    The pdf() and cdf() are exact, and are computed as a weighted sum of
    the pdf() and cdf() of the component Gaussians.

    The ppf() is computed by computing the cdf() values on a fixed
    grid and interpolating the inverse function.
    """
    # pylint: disable=protected-access

    name = 'mixmod'
    version = 0

    _support_mask = rv_continuous._support_mask

    def __init__(self, means, stds, weights, *args, **kwargs):
        """
        Create a new distribution using the given histogram

        Parameters
        ----------
        means : array_like
            The means of the Gaussians
        stds:  array_like
            The standard deviations of the Gaussians
        weights : array_like
            The weights to attach to the Gaussians
        """
        self._scipy_version_warning()
        self._means = reshape_to_pdf_size(means, -1)
        self._stds = reshape_to_pdf_size(stds, -1)
        self._weights = reshape_to_pdf_size(weights, -1)
        kwargs['shape'] = means.shape[:-1]
        self._ncomps = means.shape[-1]
        super().__init__(*args, **kwargs)
        self._addobjdata('weights', self._weights)
        self._addobjdata('stds', self._stds)
        self._addobjdata('means', self._means)

    def _scipy_version_warning(self):
        import scipy #pylint: disable=import-outside-toplevel
        scipy_version = scipy.__version__
        vtuple = scipy_version.split('.')
        if int(vtuple[0]) > 1 or int(vtuple[1]) > 7:
            return
        raise DeprecationWarning(f"Mixmod_gen will not work correctly with scipy version < 1.8.0, you have {scipy_version}")  #pragma: no cover

    @property
    def weights(self):
        """Return weights to attach to the Gaussians"""
        return self._weights

    @property
    def means(self):
        """Return means of the Gaussians"""
        return self._means

    @property
    def stds(self):
        """Return standard deviations of the Gaussians"""
        return self._stds

    def _pdf(self, x, row):
        # pylint: disable=arguments-differ
        if np.ndim(x) > 1:  #pragma: no cover
            x = np.expand_dims(x, -2)
        return (self.weights[row].swapaxes(-2,-1) *
                sps.norm(loc=self._means[row].swapaxes(-2,-1),
                         scale=self._stds[row].swapaxes(-2,-1)).pdf(x)).sum(axis=0)

    def _cdf(self, x, row):
        # pylint: disable=arguments-differ
        if np.ndim(x) > 1:  #pragma: no cover
            x = np.expand_dims(x, -2)
        return (self.weights[row].swapaxes(-2,-1) *
                sps.norm(loc=self._means[row].swapaxes(-2,-1),
                         scale=self._stds[row].swapaxes(-2,-1)).cdf(x)).sum(axis=0)

    def _ppf(self, x, row):
        # pylint: disable=arguments-differ
        min_val = np.min(self._means - 6*self._stds)
        max_val = np.max(self._means + 6*self._stds)
        grid = np.linspace(min_val, max_val, 201)
        case_idx, _, rr = get_eval_case(x, row)
        if case_idx == 1:
            cdf_vals = self.cdf(grid, rr)
        elif case_idx == 3:
            cdf_vals = self.cdf(grid, np.expand_dims(rr, -1))
        else:  #pragma: no cover
            raise ValueError(f"Opps, we handle this kind of input to mixmod._ppf {case_idx}")
        return interpolate_multi_x_y(x, row, cdf_vals, grid,
                                     bounds_error=False, fill_value=(min_val, max_val)).ravel()



    def _updated_ctor_param(self):
        """
        Set the bins as additional constructor argument
        """
        dct = super()._updated_ctor_param()
        dct['means'] = self._means
        dct['stds'] = self._stds
        dct['weights'] = self._weights
        return dct


    @classmethod
    def get_allocation_kwds(cls, npdf, **kwargs):
        """
        Return the keywords necessary to create an 'empty' hdf5 file with npdf entries
        for iterative file writeout.  We only need to allocate the objdata columns, as
        the metadata can be written when we finalize the file.
        Parameters
        ----------
        npdf: (int) number of *total* PDFs that will be written out
        kwargs: (dict) dictionary of kwargs needed to create the ensemble
        """
        if 'means' not in kwargs: #pragma: no cover
            raise ValueError("required argument means not included in kwargs")

        ncomp = np.shape(kwargs['means'])[-1]
        return dict(means=((npdf, ncomp), 'f4'), stds=((npdf, ncomp), 'f4'), weights=((npdf, ncomp), 'f4'))


    @classmethod
    def add_mappings(cls):
        """
        Add this classes mappings to the conversion dictionary
        """
        cls._add_creation_method(cls.create, None)
        cls._add_extraction_method(extract_mixmod_fit_samples, None)

    @classmethod
    def make_test_data(cls):
        """ Make data for unit tests """
        cls.test_data = dict(mixmod=dict(gen_func=mixmod,\
                                         ctor_data=dict(weights=WEIGHT_MIXMOD,\
                                                        means=MEAN_MIXMOD,\
                                                        stds=STD_MIXMOD),\
                                         convert_data={}, test_xvals=TEST_XVALS,
                                         atol_diff2=1.))


mixmod = mixmod_gen.create

add_class(mixmod_gen)



"""
Unit tests for PDF class
"""
import sys
import unittest
from functools import partial
import qp

from qp import test_funcs


class PDFTestCase(unittest.TestCase):
    """ Class to manage automatically generated tests for qp distributions """

    def setUp(self):
        """
        Make any objects that are used in multiple tests.
        """

    def tearDown(self):
        "Clean up any mock data files created by the tests."


    @classmethod
    def auto_add_class(cls, test_class, ens_list):
        """Add tests as member functions to a class"""
        for key, val in test_class.test_data.items():
            test_pdf = val.pop('test_pdf', True)
            if test_pdf:
                kw_test_pdf = dict(short=val.pop('short', False), check_props=val.pop('check_props', True))
                the_pdf_func = partial(test_funcs.run_pdf_func_tests, test_class, val, **kw_test_pdf)
                setattr(cls, 'test_pdf_%s' % key, the_pdf_func)
            test_persist = val.pop('test_persist', True)
            if test_persist:
                the_persist_func = partial(test_funcs.run_persist_func_tests, val)
                setattr(cls, 'test_persist_%s' % key, the_persist_func)
            test_convert = val.pop('test_convert', True)
            if 'convert_data' not in val:
                test_convert = False
            if test_convert:
                kw_test_convert = dict(atol_diff=val.pop('atol_diff', 1e-2), atol_diff2=val.pop('atol_diff2', 1e-2))
                for i, ens in enumerate(ens_list):
                    the_convert_func = partial(test_funcs.run_convert_tests, ens_orig=ens, gen_class=test_class, test_data=val, **kw_test_convert)
                    setattr(cls, 'test_convert_%s_%i' % (key, i), the_convert_func)
            test_plot = val.pop('test_plot', True)
            if test_plot:
                kw_test_plot = dict(do_samples=val.pop('do_samples', False))
                the_plot_func = partial(test_funcs.run_plotting_func_tests, val, **kw_test_plot)
                setattr(cls, 'test_plotting_%s' % key, the_plot_func)

    @classmethod
    def auto_add(cls, class_list, ens_orig):
        """Add tests as member functions to a class"""
        for test_class in class_list:
            try:
                test_class.make_test_data()
            except AttributeError:
                pass
            if hasattr(test_class, 'test_data'):
                cls.auto_add_class(test_class, ens_orig)


ENS_ORIG = test_funcs.build_ensemble(qp.stats.norm_gen.test_data['norm'])  #pylint: disable=no-member
ENS_MULTI = test_funcs.build_ensemble(qp.stats.norm_gen.test_data['norm'])  #pylint: disable=no-member
TEST_CLASSES = qp.instance().values()

PDFTestCase.auto_add(TEST_CLASSES, [ENS_ORIG, ENS_MULTI])


if __name__ == '__main__':
    unittest.main()



"""This module implements functions to convert distributions between various representations
These functions should then be registered with the `qp.ConversionDict` using `qp_add_mapping`.
That will allow the automated conversion mechanisms to work.
"""
import numpy as np

from sklearn import mixture
from scipy import integrate as sciint
from scipy import interpolate as sciinterp

from .sparse_rep import indices2shapes, build_sparse_representation, decode_sparse_indices

def extract_vals_at_x(in_dist, **kwargs):
    """Convert using a set of x and y values.

    Parameters
    ----------
    in_dist : `qp.Ensemble`
        Input distributions

    Keywords
    --------
    xvals : `np.array`
        Locations at which the pdf is evaluated

    Returns
    -------
    data : `dict`
        The extracted data
    """
    xvals = kwargs.pop('xvals', None)
    if xvals is None: # pragma: no cover
        raise ValueError("To convert to extract_xy_vals you must specify xvals")
    yvals = in_dist.pdf(xvals)
    return dict(xvals=xvals, yvals=yvals)


def extract_xy_vals(in_dist, **kwargs):
    """Convert using a set of x and y values.

    Parameters
    ----------
    in_dist : `qp.Ensemble`
        Input distributions

    Keywords
    --------
    xvals : `np.array`
        Locations at which the pdf is evaluated

    Returns
    -------
    data : `dict`
        The extracted data
    """
    xvals = kwargs.pop('xvals', None)
    if xvals is None: # pragma: no cover
        raise ValueError("To convert using extract_xy_vals you must specify xvals")
    yvals = in_dist.pdf(xvals)
    expand_x = np.ones(yvals.shape) * np.squeeze(xvals)
    return dict(xvals=expand_x, yvals=yvals)


def extract_samples(in_dist, **kwargs):
    """Convert using a set of values sampled from the PDF

    Parameters
    ----------
    in_dist : `qp.Ensemble`
        Input distributions

    Keywords
    --------
    size : `int`
        Number of samples to generate

    Returns
    -------
    data : `dict`
        The extracted data
    """
    samples = in_dist.rvs(size=kwargs.pop('size', 1000))
    xvals = kwargs.pop('xvals')
    return dict(samples=samples, xvals=xvals, yvals=None)


def extract_hist_values(in_dist, **kwargs):
    """Convert using a set of values sampled from the PDF

    Parameters
    ----------
    in_dist : `qp.Ensemble`
        Input distributions

    Keywords
    --------
    bins : `np.array`
        Histogram bin edges

    Returns
    -------
    data : `dict`
        The extracted data
    """
    bins = kwargs.pop('bins', None)
    if bins is None: # pragma: no cover
        raise ValueError("To convert using extract_hist_samples you must specify bins")
    bins, pdfs = in_dist.histogramize(bins)
    return dict(bins=bins, pdfs=pdfs)


def extract_hist_samples(in_dist, **kwargs):
    """Convert using a set of values samples that are then histogramed

    Parameters
    ----------
    in_dist : `qp.Ensemble`
        Input distributions

    Keywords
    --------
    bins : `np.array`
        Histogram bin edges
    size : `int`
        Number of samples to generate

    Returns
    -------
    data : `dict`
        The extracted data
    """
    bins = kwargs.pop('bins', None)
    size = kwargs.pop('size', 1000)
    if bins is None: # pragma: no cover
        raise ValueError("To convert using extract_hist_samples you must specify bins")
    samples = in_dist.rvs(size=size)

    def hist_helper(sample):
        return np.histogram(sample, bins=bins)[0]
    vv = np.vectorize(hist_helper, signature="(%i)->(%i)" % (samples.shape[0], bins.size-1))
    pdfs = vv(samples)
    return dict(bins=bins, pdfs=pdfs)


def extract_quantiles(in_dist, **kwargs):
    """Convert using a set of quantiles and the locations at which they are reached

    Parameters
    ----------
    in_dist : `qp.Ensemble`
        Input distributions

    Keywords
    --------
    quantiles : `np.array`
        Quantile values to use

    Returns
    -------
    data : `dict`
        The extracted data
    """
    quants = kwargs.pop('quants', None)
    if quants is None: # pragma: no cover
        raise ValueError("To convert using extract_quantiles you must specify quants")
    locs = in_dist.ppf(quants)
    return dict(quants=quants, locs=locs)


def extract_fit(in_dist, **kwargs): # pragma: no cover
    """Convert to a functional distribution by fitting it to a set of x and y values

    Parameters
    ----------
    in_dist : `qp.Ensemble`
        Input distributions

    Keywords
    --------
    xvals : `np.array`
        Locations at which the pdf is evaluated

    Returns
    -------
    data : `dict`
        The extracted data
    """
    raise NotImplementedError('extract_fit')
    #xvals = kwargs.pop('xvals', None)
    #if xvals is None:
    #   raise ValueError("To convert using extract_fit you must specify xvals")
    ##vals = in_dist.pdf(xvals)


def extract_mixmod_fit_samples(in_dist, **kwargs):
    """Convert to a mixture model using a set of values sample from the pdf

    Parameters
    ----------
    in_dist : `qp.Ensemble`
        Input distributions

    Keywords
    --------
    ncomps : `int`
        Number of components in mixture model to use
    nsamples : `int`
        Number of samples to generate
    random_state : `int`
        Used to reproducibly generate random variate from in_dist

    Returns
    -------
    data : `dict`
        The extracted data
    """
    n_comps = kwargs.pop('ncomps', 3)
    n_sample = kwargs.pop('nsamples', 1000)
    random_state = kwargs.pop('random_state', None)
    samples = in_dist.rvs(size=n_sample, random_state=random_state)
    def mixmod_helper(samps):
        estimator = mixture.GaussianMixture(n_components=n_comps)
        estimator.fit(samps.reshape(-1, 1))
        weights = estimator.weights_
        means = estimator.means_[:, 0]
        stdevs = np.sqrt(estimator.covariances_[:, 0, 0])
        ov = np.vstack([weights, means, stdevs])
        return ov

    vv = np.vectorize(mixmod_helper, signature="(%i)->(3,%i)" % (n_sample, n_comps))
    fit_vals = vv(samples)
    return dict(weights=fit_vals[:, 0, :], means=fit_vals[:, 1, :], stds=fit_vals[:, 2, :])

def extract_voigt_mixmod(in_dist, **kwargs): #pragma: no cover
    """Convert to a voigt mixture model starting with a gaussian mixture model,
    trivially by setting gammas to 0

    Parameters
    ----------
    in_dist : `qp.Ensemble`
        Input distributions

    Returns
    -------
    data : `dict`
        The extracted data
    """
    objdata = in_dist.objdata()
    means = objdata['means']
    stds = objdata['stds']
    weights = objdata['weights']
    gammas = np.zeros_like(means)
    return dict(means=means, stds=stds, weights=weights, gammas=gammas, **kwargs)


def extract_voigt_xy(in_dist, **kwargs): #pragma: no cover
    """Build a voigt function basis and run a match-pursuit algorithm to fit gridded data

    Parameters
    ----------
    in_dist : `qp.Ensemble`
        Input distributions

    Returns
    -------
    data : `dict`
        The extracted data as sparse indices, basis, and metadata to rebuild the basis
    """

    sparse_results = extract_voigt_xy_sparse(in_dist, **kwargs)
    indices = sparse_results['indices']
    meta = sparse_results['metadata']

    w, m, s, g = indices2shapes(indices, meta)
    return dict(means=m, stds=s, weights=w, gammas=g)


def extract_voigt_xy_sparse(in_dist, **kwargs): #pragma: no cover
    """Build a voigt function basis and run a match-pursuit algorithm to fit gridded data

    Parameters
    ----------
    in_dist : `qp.Ensemble`
        Input distributions

    Returns
    -------
    data : `dict`
        The extracted data as shaped parameters means, stds, weights, gammas
    """

    yvals = in_dist.objdata()['yvals']

    default = in_dist.metadata()['xvals'][0]
    z = kwargs.pop('xvals', default)
    nz = kwargs.pop('nz', 300)

    minz = np.min(z)
    _, j = np.where(yvals > 0)
    maxz = np.max(z[j])
    newz = np.linspace(minz, maxz, nz)
    interp = sciinterp.interp1d(z, yvals, assume_sorted=True)
    newpdf = interp(newz)
    newpdf = newpdf / sciint.trapz(newpdf, newz).reshape(-1, 1)
    ALL, bigD, _ = build_sparse_representation(newz, newpdf)
    return dict(indices=ALL, metadata=bigD)

def extract_sparse_from_xy(in_dist, **kwargs): #pragma: no cover
    """Extract sparse representation from an xy interpolated representation

    Parameters
    ----------
    in_dist : `qp.Ensemble`
        Input distributions

    Keywords
    --------
    xvals : array-like
        Used to override the y-values
    xvals : array-like
        Used to override the x-values
    nvals : int
        Used to override the number of bins

    Returns
    -------
    metadata : `dict`
        Dictionary with data for sparse representation

    Notes
    -----
    This function will rebin to a grid more suited to the in_dist support by
    removing x-values corrsponding to y=0
    """
    default = in_dist.objdata()['yvals']
    yvals = kwargs.pop('yvals', default)
    default = in_dist.metadata()['xvals'][0]
    xvals = kwargs.pop('xvals', default)
    nvals = kwargs.pop('nvals', 300)
    #rebin to a grid more suited to the in_dist support
    xmin = np.min(xvals)
    _, j = np.where(yvals > 0)
    xmax = np.max(xvals[j])
    newx = np.linspace(xmin, xmax, nvals)
    interp = sciinterp.interp1d(xvals, yvals, assume_sorted=True)
    newpdf = interp(newx)
    sparse_indices, metadata, _ = build_sparse_representation(newx, newpdf)
    metadata['xvals'] = newx
    metadata['sparse_indices'] = sparse_indices
    metadata.pop('Ntot')
    return metadata


def extract_xy_sparse(in_dist, **kwargs): #pragma: no cover
    """Extract xy-interpolated representation from an sparese representation

    Parameters
    ----------
    in_dist : `qp.Ensemble`
        Input distributions

    Keywords
    --------
    xvals : array-like
        Used to override the y-values
    xvals : array-like
        Used to override the x-values
    nvals : int
        Used to override the number of bins

    Returns
    -------
    metadata : `dict`
        Dictionary with data for interpolated representation

    Notes
    -----
    This function will rebin to a grid more suited to the in_dist support by
    removing x-values corrsponding to y=0
    """

    yvals = in_dist.objdata()['yvals']
    default = in_dist.metadata()['xvals'][0]
    xvals = kwargs.pop('xvals', default)
    nvals = kwargs.pop('nvals', 300)
    #rebin to a grid more suited to the in_dist support
    xmin = np.min(xvals)
    _, j = np.where(yvals > 0)
    xmax = np.max(xvals[j])
    newx = np.linspace(xmin, xmax, nvals)
    interp = sciinterp.interp1d(xvals, yvals, assume_sorted=True)
    newpdf = interp(newx)
    sparse_indices, sparse_meta, A = build_sparse_representation(newx, newpdf)
    #decode the sparse indices into basis indices and weights
    basis_indices, weights = decode_sparse_indices(sparse_indices)
    #retrieve the weighted array of basis functions for each object
    pdf_y = A[:, basis_indices] * weights
    #normalize and sum the weighted pdfs
    x = sparse_meta['z']
    y = pdf_y.sum(axis=-1)
    norms = sciint.trapz(y.T, x)
    y /= norms
    #super(sparse_gen, self).__init__(x, y.T, *args, **kwargs)
    xvals = x
    yvals = y.T
    return dict(xvals=xvals, yvals=yvals, **kwargs)



"""This module implements some performance metrics for distribution parameterization"""
import logging
from collections import namedtuple
from functools import partial
from deprecated import deprecated

import numpy as np

from qp.metrics import array_metrics
from qp.metrics.goodness_of_fit import goodness_of_fit_metrics
from qp.metrics.brier import Brier
from qp.utils import epsilon

Grid = namedtuple('Grid', ['grid_values', 'cardinality', 'resolution', 'hist_bin_edges', 'limits'])

def _calculate_grid_parameters(limits, dx:float=0.01) -> Grid:
    """
    Create a grid of points and return parameters describing it.

    Args:
        limits (Iterable) often a 2-tuple or numpy array with shape (2,)
            the max and min values of the 1d grid
        dx (float, optional):
            the desired delta between points. Used to define the cardinality. Defaults to 0.01.

    Returns:
        Grid: a namedtuple containing a 1d grid's values and attributes.
            grid_values: np.array with size = cardinality
            cardinality: int, number of elements in grid_value
            resolution: float, equal to grid_values[i] - grid_values[i-1]
            hist_bin_edges: np.array with size = cardinality+1.
                Equally spaced histogram bin edges starting at limit-resolution/2.
                Assumes that grid_value[i] should be centered in the bin defined by
                (hist_bin_edge[i], hist_bin_edge[i+1]).
            limits: 2-tuple, the limits passed in and used in this function
    """
    cardinality = int((limits[-1] - limits[0]) / dx)
    grid_values = np.linspace(limits[0], limits[1], cardinality)
    resolution = (limits[-1] - limits[0]) / (cardinality - 1)
    hist_bin_edges = np.histogram_bin_edges((limits[0]-resolution/2, limits[1]+resolution/2), cardinality)

    return Grid(grid_values, cardinality, resolution, hist_bin_edges, limits)

def calculate_moment(p, N, limits, dx=0.01):
    """
    Calculates a moment of a qp.Ensemble object

    Parameters
    ----------
    p: qp.Ensemble object
        the collection of PDFs whose moment will be calculated
    N: int
        order of the moment to be calculated
    limits: tuple of floats
        endpoints of integration interval over which to calculate moments
    dx: float
        resolution of integration grid

    Returns
    -------
    M: float
        value of the moment
    """
    # Make a grid from the limits and resolution
    grid = _calculate_grid_parameters(limits, dx)

    # Evaluate the functions on the grid
    pe = p.gridded(grid.grid_values)[1]

    # calculate the moment
    grid_to_N = grid.grid_values ** N
    M = array_metrics.quick_moment(pe, grid_to_N, grid.resolution)

    return M


def calculate_kld(p, q, limits, dx=0.01):
    """
    Calculates the Kullback-Leibler Divergence between two qp.Ensemble objects.

    Parameters
    ----------
    p: Ensemble object
        probability distribution whose distance _from_ `q` will be calculated.
    q: Ensemble object
        probability distribution whose distance _to_ `p` will be calculated.
    limits: tuple of floats
        endpoints of integration interval in which to calculate KLD
    dx: float
        resolution of integration grid

    Returns
    -------
    Dpq: float
        the value of the Kullback-Leibler Divergence from `q` to `p`

    Notes
    -----
    TO DO: have this take number of points not dx!
    """
    if p.shape != q.shape:
        raise ValueError('Cannot calculate KLD between two ensembles with different shapes')

    # Make a grid from the limits and resolution
    grid = _calculate_grid_parameters(limits, dx)

    # Evaluate the functions on the grid and normalize
    pe = p.gridded(grid.grid_values)
    pn = pe[1]
    qe = q.gridded(grid.grid_values)
    qn = qe[1]

    # Calculate the KLD from q to p
    Dpq = array_metrics.quick_kld(pn, qn, grid.resolution)# np.dot(pn * logquotient, np.ones(len(grid)) * dx)

    if np.any(Dpq < 0.): #pragma: no cover
        print('broken KLD: '+str((Dpq, pn, qn, grid.resolution)))
        Dpq = epsilon*np.ones(Dpq.shape)
    return Dpq


def calculate_rmse(p, q, limits, dx=0.01):
    """
    Calculates the Root Mean Square Error between two qp.Ensemble objects.

    Parameters
    ----------
    p: qp.Ensemble object
        probability distribution function whose distance between its truth and the approximation of `q` will be calculated.
    q: qp.Ensemble object
        probability distribution function whose distance between its approximation and the truth of `p` will be calculated.
    limits: tuple of floats
        endpoints of integration interval in which to calculate RMS
    dx: float
        resolution of integration grid

    Returns
    -------
    rms: float
        the value of the RMS error between `q` and `p`

    Notes
    -----
    TO DO: change dx to N
    """
    if p.shape != q.shape:
        raise ValueError('Cannot calculate RMSE between two ensembles with different shapes')

    # Make a grid from the limits and resolution
    grid = _calculate_grid_parameters(limits, dx)

    # Evaluate the functions on the grid
    pe = p.gridded(grid.grid_values)[1]
    qe = q.gridded(grid.grid_values)[1]

    # Calculate the RMS between p and q
    rms = array_metrics.quick_rmse(pe, qe, grid.cardinality)# np.sqrt(dx * np.sum((pe - qe) ** 2))

    return rms


def calculate_rbpe(p, limits=(np.inf, np.inf)):
    """
    Calculates the risk based point estimates of a qp.Ensemble object.
    Algorithm as defined in 4.2 of 'Photometric redshifts for Hyper Suprime-Cam 
    Subaru Strategic Program Data Release 1' (Tanaka et al. 2018).

    Parameters
    ----------
    p: qp.Ensemble object
        Ensemble of PDFs to be evalutated
    limits: tuple
        The limits at which to evaluate possible z_best estimates.
        If custom limits are not provided then all potential z value will be
        considered using the scipy.optimize.minimize_scalar function.

    Returns
    -------
    rbpes: array of floats
        The risk based point estimates of the provided ensemble.
    """
    rbpes = []

    def evaluate_pdf_at_z(z, dist):
        return dist.pdf(z)[0][0]

    for n in range(0, p.npdf):

        if p[n].npdf != 1:
            raise ValueError('quick_rbpe only handles Ensembles with a single PDF '
                             'for ensembles with more than one PDF, use the qp.metrics.risk_based_point_estimate function.')

        this_dist_pdf_at_z = partial(evaluate_pdf_at_z, dist=p[n])
        integration_bounds = (p[n].ppf(0.01)[0][0], p[n].ppf(0.99)[0][0])

        rbpes.append(array_metrics.quick_rbpe(this_dist_pdf_at_z, integration_bounds, limits))

    return np.array(rbpes)

def calculate_brier(p, truth, limits, dx=0.01):
    """This function will do the following:
    1) Generate a Mx1 sized grid based on `limits` and `dx`.
    2) Produce an NxM array by evaluating the pdf for each of the N distribution objects in the Ensemble p on the grid. 
    3) Produce an NxM truth_array using the input truth and the generated grid. All values will be 0 or 1.
    4) Create a Brier metric evaluation object
    5) Return the result of the Brier metric calculation.

    Parameters
    ----------
    p: qp.Ensemble object 
        of N distributions probability distribution functions that will be gridded and compared against truth.
    truth: Nx1 sequence
        the list of true values, 1 per distribution in p.
    limits: 2-tuple of floats
        endpoints grid to evaluate the PDFs for the distributions in p
    dx: float
        resolution of the grid Defaults to 0.01.

    Returns
    -------
    Brier_metric: float
    """

    # Ensure that the number of distributions objects in the Ensemble is equal to the length of the truth array
    if p.npdf != len(truth):
        raise ValueError("Number of distributions in the Ensemble (%d) is not equal to the number of truth values (%d)" % (p.npdf, len(truth)))

    # Values of truth that are outside the defined limits will not appear truth_array.
    # Consider expanding the limits or using numpy.clip to restrict truth values to the limits.
    if np.any(np.less(truth, limits[0])) or np.any(np.greater(truth, limits[1])):
        raise ValueError("Input truth values exceed the defined limits")

    # Make a grid object that defines grid values and histogram bin edges using limits and dx
    grid = _calculate_grid_parameters(limits, dx)

    # Evaluate the pdf of the distributions on the grid.
    # The value returned from p.gridded is a 2-tuple. The 0th index is the array of grid points,
    # the 1st index is the array of PDF values. Thus we call p.gridded(...)[1]
    pdf_values = p.gridded(grid.grid_values)[1]

    # Create the NxM truth_array.
    # Note np.histogram returns a 2-tuple. The 0th index is the histogram array,
    # thus we call np.squeeze to remove extra dimensions.
    truth_array = np.squeeze([np.histogram(t, grid.hist_bin_edges)[0] for t in truth])

    # instantiate the Brier metric object
    brier_metric_evaluation = Brier(pdf_values, truth_array)

    # return the results of evaluating the Brier metric
    return brier_metric_evaluation.evaluate()

@deprecated(
    reason="""
    This implementation is deprecated for performance reasons and does not accommodate N vs 1 comparisons.
    Please use calculate_goodness_of_fit instead. This method is for testing purposes only.
    """,
    category=DeprecationWarning)
def calculate_anderson_darling(p, scipy_distribution='norm', num_samples=100, _random_state=None):  # pylint: disable=unused-argument
    """This function is deprecated and will be completely removed in a later version.
    Please use `calculate_goodness_of_fit` instead.

    Returns
    -------
    logger.warning
    """
    logging.warning("This function is deprecated, please use `calculate_goodness_of_fit`") # pragma: no cover

@deprecated(
    reason="""
    This implementation is deprecated for performance reasons and does not accommodate N vs 1 comparisons.
    Please use calculate_goodness_of_fit instead. This method is for testing purposes only.
    """,
    category=DeprecationWarning)
def calculate_cramer_von_mises(p, q, num_samples=100, _random_state=None, **kwargs):  # pylint: disable=unused-argument
    """This function is deprecated and will be completely removed in a later version.
    Please use `calculate_goodness_of_fit` instead.

    Returns
    -------
    logger.warning
    """
    logging.warning("This function is deprecated, please use `calculate_goodness_of_fit`") # pragma: no cover

@deprecated(
    reason="""
    This implementation is deprecated for performance reasons and does not accommodate N vs 1 comparisons.
    Please use calculate_goodness_of_fit instead. This method is for testing purposes only.
    """,
    category=DeprecationWarning)
def calculate_kolmogorov_smirnov(p, q, num_samples=100, _random_state=None):  # pylint: disable=unused-argument
    """This function is deprecated and will be completely removed in a later version.
    Please use `calculate_goodness_of_fit` instead.

    Returns
    -------
    logger.warning
    """
    logging.warning("This function is deprecated, please use `calculate_goodness_of_fit`") # pragma: no cover

def calculate_outlier_rate(p, lower_limit=0.0001, upper_limit=0.9999):
    """Fraction of outliers in each distribution

    Parameters
    ----------
    p : qp.Ensemble
        A collection of N distributions. This implementation expects that Ensembles are not nested.
    lower_limit : float, optional
        Lower bound for outliers, by default 0.0001
    upper_limit : float, optional
        Upper bound for outliers, by default 0.9999

    Returns
    -------
    [float]
        1xN array where each element is the percent of outliers for a distribution in the Ensemble.
    """

    # Validate that all the distributions in the Ensemble are single distributions - i.e. no nested Ensembles
    try:
        _check_ensemble_is_not_nested(p)
    except ValueError:  #pragma: no cover - unittest coverage for _check_ensemble_is_not_nested is complete
        logging.warning("Each element in the ensemble `p` must be a single distribution.")

    outlier_rates = [(dist.cdf(lower_limit) + (1. - dist.cdf(upper_limit)))[0][0] for dist in p]
    return outlier_rates

def calculate_goodness_of_fit(estimate, reference, fit_metric='ad', num_samples=100, _random_state=None):
    """This method calculates goodness of fit between the distributions in the
    `estimate` and `reference` Ensembles using the specified fit_metric.

    Parameters
    ----------
    estimate : Ensemble containing N distributions
        Random variate samples will be drawn from this Ensemble
    reference : Ensemble containing N or 1 distributions
        The CDF of the distributions in this Ensemble are used in the goodness of fit
        calculation.
    fit_metric : string, optional
        The goodness of fit metric to use. One of ['ad', 'cvm', 'ks']. For clarity,
        'ad' = Anderson-Darling, 'cvm' = Cramer-von Mises, and 'ks' = Kolmogorov-Smirnov, by default 'ad'
    num_samples : int, optional
        Number of random variates to draw from each distribution in `estimate`, by default 100
    _random_state : _type_, optional
        Used for testing to create reproducible sets of random variates, by default None

    Returns
    -------
    output: [float]
        A array of floats where each element is the result of the statistic calculation.

    Raises
    ------
    KeyError
        If the requested `fit_metric` is not contained in `goodness_of_fit_metrics` dictionary,
        raise a KeyError.

    Notes
    -----
    The calculation of the goodness of fit metrics is not symmetric.
    i.e. `calculate_goodness_of_fit(p, q, ...) != calculate_goodness_of_fit(q, p, ...)`

    In the future, we should be able to do this directly from the PDFs without needing to
    take random variates from the `estimate` Ensemble.

    The vectorized implementations of fit metrics are copied over (unmodified) from
    the developer branch of Scipy 1.10.0dev. When Scipy 1.10 is released, we can replace
    the copied implementation with the ones in Scipy.
    """

    try:
        _check_ensembles_contain_correct_number_of_distributions(estimate, reference)
    except ValueError: #pragma: no cover - unittest coverage for _check_ensembles_contain_correct_number_of_distributions is complete
        logging.warning("The ensemble `reference` should have 1 or N distributions. With N = number of distributions in the ensemble `estimate`.")

    try:
        _check_ensemble_is_not_nested(estimate)
    except ValueError:  #pragma: no cover - unittest coverage for _check_ensemble_is_not_nested is complete
        logging.warning("Each element in the ensemble `estimate` must be a single distribution.")

    try:
        _check_ensemble_is_not_nested(reference)
    except ValueError:  #pragma: no cover - unittest coverage for _check_ensemble_is_not_nested is complete
        logging.warning("Each element in the ensemble `reference` must be a single distribution.")

    if fit_metric not in goodness_of_fit_metrics:
        metrics = list(goodness_of_fit_metrics.keys())
        raise KeyError(f"`fit_metric` should be one of {metrics}.")

    return goodness_of_fit_metrics[fit_metric](
        reference,
        np.squeeze(estimate.rvs(size=num_samples, random_state=_random_state))
    )

def _check_ensembles_are_same_size(p, q):
    """This utility function ensures checks that two Ensembles contain an equal number of distribution objects.

    Args:
        p qp.Ensemble: An Ensemble containing 0 or more distributions
        q qp.Ensemble: A second Ensemble containing 0 or more distributions

    Raises:
        ValueError: If the result of evaluating qp.Ensemble.npdf on each Ensemble is not the same, raise an error.
    """
    if p.npdf != q.npdf:
        raise ValueError("Input ensembles should have the same number of distributions")

def _check_ensemble_is_not_nested(p):
    """This utility function ensures that each element in the Ensemble is a single distribution.

    Args:
        p qp.Ensemble: An Ensemble that could contain nested Ensembles with multiple distributions in each

    Raises:
        ValueError: If there are any elements of the input Ensemble that contain more than 1 PDF, raise an error.
    """
    for dist in p:
        if dist.npdf != 1:
            raise ValueError("Each element in the input Ensemble should be a single distribution.")

def _check_ensembles_contain_correct_number_of_distributions(estimate, reference):
    """This utility function ensures that the number of distributions in the ensembles matches
    expectations. estimate can contain 1 to N distributions.
    reference should contain either 1 or the same number of distributions
    as estimate.

    Example logic:
    estimate=N, reference=N (1 <= N) -> Pass
    estimate=N, reference=1 (1 <= N) -> Pass
    estimate=1, reference=N (N != 1) -> Raise ValueError
    estimate=N, reference=M (N != M) -> Raise ValueError

    Parameters
    ----------
    estimate : Ensemble
        Used to calculate goodness of fit metrics, random
        variates will be produced from the distributions in this ensemble.
    reference : Ensemble
        The CDFs of the distributions in this ensemble will be used to calculate
        goodness of fit metrics.

    Raises
    ------
    ValueError
        If the number of distributions in the reference ensemble does not meet the requirements,
        raise a ValueError.
    """

    if estimate.npdf == reference.npdf:
        pass
    elif reference.npdf == 1:
        pass
    else:
        raise ValueError("`reference` should contain either 1 distribution or the same number of distributions as `estimate`.")



from typing import List

import numpy as np
from scipy.interpolate import interp1d

from qp.quantile_pdf_constructors.abstract_pdf_constructor import AbstractQuantilePdfConstructor

class DualSplineAverage(AbstractQuantilePdfConstructor):
    """Implementation of the "area-under-the-curve" using the average of
    the bounding splines fit to the CDF derivative.

    By using the difference between quantiles to solve for the area under the PDF, we can
    create an approximation of the original PDF. However, because we use a piecewise linear
    approximation for the continuous PDF, our approximated p(z) values will always be different
    that the original distribution. In practice they typically oscillate above and below the
    original curve as each calculation attempts to correct for over or undershooting of the prior
    calculation.

    If we fit two splines, one to the odd and one to the even approximated points, then take the
    average, the resulting average of those splines tend to fit the original distribution well.

    This constructor implements that algorithmic approach.
    """

    def __init__(self, quantiles:List[float], locations: List[List[float]]) -> None:
        """Constructor to instantiate this class.

        Parameters
        ----------
        quantiles : List[float]
            List of n quantile values in the range (0,1).
        locations : List[List[float]]
            List of m Lists, each containing n values corresponding to the
            y-value of the PPF function at the same quantile index.
        """
        self._quantiles = quantiles
        self._locations = np.atleast_2d(locations)

        self._p_of_zs = None
        self.y1 = None
        self.y2 = None

    def prepare_constructor(self) -> None:
        """This method solves for the area under the PDF via a stepwise algorithm.
        Given that the difference between any two quantile values is equal to the
        area under the PDF between the corresponding pair of locations, _and_ given
        that we know the p(z) value at 1 of those locations, we can solve
        for the unknown p(z) value at the other location.

        We approximate the area under the curve as a trapezoid with the following area:
        (q_i+1 - q_i) = (loc_i+1 - loc_i) * p(z_i) + (1/2) * (loc_i+1 - loc_i) * p(z_i+1 = p(z_i))

        Solving for p(z_i+1), we have:
        p(z_i+1) = [2 * (q_i+1 - q_i) / (loc_i+1 - loc_i)] - p(z_i)

        The first term in this equation is calculated as `first_term`. After that we step along
        all distributions simultaneously for each location, using the previous p(z) value to
        calculate the next.
        """

        # Prepare an empty container for the output
        self._p_of_zs = np.zeros(self._locations.shape)

        # Prepare an all-zero list, used in the step-wise calculation to prevent negative values
        zeros_for_comparison = np.zeros(self._locations.shape[0])

        # Calculate the first term
        first_term = 2 * np.diff(self._quantiles) / np.diff(self._locations)

        # Perform the step-wise calculation for all distributions simultaneously
        for i in range(1, np.shape(self._p_of_zs)[-1]):
            self._p_of_zs[:,i] = np.maximum(zeros_for_comparison, first_term[:,i-1] - self._p_of_zs[:,i-1])

        # Set any negative values to 0.
        self._p_of_zs = np.maximum(np.zeros(self._locations.shape), self._p_of_zs)

    def construct_pdf(self, grid: List[float], row: List[int] = None) -> List[List[float]]:
        """This method utilizes intermediate calculations from `prepare_constructor`
        along with the provided grid (i.e. x) values to return corresponding y values
        to construct the PDF approximation.

        Parameters
        ----------
        grid : List[float]
            x values used to calculate corresponding y values
        row : List[int], optional
            A list of indexes of the original distribution to return, used as a filter.
            By default None will do no filtering.

        Returns
        -------
        List[List[float]]
            The lists of y values returned from self._interpolation_functions
        """
        if self._p_of_zs is None:
            self.prepare_constructor()

        # Support the use of `row` as a filter. If row is None, do nothing,
        # otherwise, return a subset of the rows.
        # Using `map` alone will return an iterator that will be completely consumed after the first
        # list comprehension. Thus we convert the map to a list so that it can be used multiple times.
        filtered_p_of_zs = self._p_of_zs
        filtered_locations = self._locations
        if row is not None:
            filtered_p_of_zs = list(map(self._p_of_zs.__getitem__, np.unique(row)))
            filtered_locations = list(map(self._locations.__getitem__, np.unique(row)))

        # Create a list of interpolated splines for the even and odd pairs of (specific_locations, specific_p_of_zs)
        f1 = np.asarray(
            [
                interp1d(np.squeeze(specific_locations[0::2]), np.squeeze(specific_p_of_zs[0::2]), bounds_error=False, fill_value=0.0, kind='cubic')
                for specific_p_of_zs, specific_locations in zip(filtered_p_of_zs, filtered_locations)
            ]
        )

        f2 = np.asarray(
            [
                interp1d(np.squeeze(specific_locations[1::2]), np.squeeze(specific_p_of_zs[1::2]), bounds_error=False, fill_value=0.0, kind='cubic')
                for specific_p_of_zs, specific_locations in zip(filtered_p_of_zs, filtered_locations)
            ]
        )

        # Evaluate all the splines at the input grid values
        self.y1 = np.asarray([func_1(grid) for func_1 in f1])
        self.y2 = np.asarray([func_2(grid) for func_2 in f2])

        # Return the average of the spline values at each of the evaluated points.
        return (self.y1 + self.y2) / 2

    def debug(self):
        """This is a debugging utility that is meant to return intermediate calculation values
        to make it easier to visualize and debug the reconstruction algorithm.

        Returns
        -------
            _quantiles :
                Input during constructor instantiation
            _locations :
                Input during constructor instantiation
            _p_of_zs :
                Resulting p(z) values found after calculating the area of trapezoids based on the difference between adjacent quantile values
            y1 :
                One of two splines fit to alternating pairs of (_locations, _p_of_zs)
            y2 :
                One of two splines fit to alternating pairs of (_locations, _p_of_zs)
        """
        return self._quantiles, self._locations, self._p_of_zs, self.y1, self.y2



import unittest
import numpy as np
import qp
from qp import interp_gen
from qp.ensemble import Ensemble
from qp.metrics.pit import PIT


# constants for tests
NMAX = 2.5
NPDF = 399
ZGRID = np.linspace(0, NMAX, 301)

ADVAL_ALL = 82.51480
ADVAL_CUT = 1.10750
CVMVAL = 20.63155
KSVAL = 0.367384
OUTRATE = 0.0

# These constants retained for future use
# CDEVAL = -4.31200
# SIGIQR = 0.0045947
# BIAS = -0.00001576
# SIGMAD = 0.0046489

class PitTestCase(unittest.TestCase):
    """ Test cases for PIT metric. """

    def setUp(self):
        np.random.seed(87)
        self.true_zs = np.random.uniform(high=NMAX, size=NPDF)

        locs = np.expand_dims(self.true_zs + np.random.normal(0.0, 0.01, NPDF), -1)
        scales = np.ones((NPDF, 1)) * 0.1 + np.random.uniform(size=(NPDF, 1)) * .05
        self.n_ens = Ensemble(qp.stats.norm, data=dict(loc=locs, scale=scales))

        self.grid_ens = self.n_ens.convert_to(interp_gen, xvals=ZGRID)

    def test_pit_metrics(self):
        """Base test of PIT metric generation"""
        quant_grid = np.linspace(0, 1, 101)
        pit_obj = PIT(self.grid_ens, self.true_zs, quant_grid)

        pit_samples = pit_obj.pit_samps
        self.assertTrue(len(pit_samples) == 399)

        pit_ensemble = pit_obj.pit
        self.assertTrue(pit_ensemble.npdf == 1)

        meta_metrics = pit_obj.calculate_pit_meta_metrics()

        ad_stat = meta_metrics['ad'].statistic
        assert np.isclose(ad_stat, ADVAL_ALL)

        cut_ad_stat = pit_obj.evaluate_PIT_anderson_ksamp(pit_min=0.6, pit_max=0.9).statistic
        assert np.isclose(cut_ad_stat, ADVAL_CUT)

        cvm_stat = meta_metrics['cvm'].statistic
        assert np.isclose(cvm_stat, CVMVAL)

        ks_stat = meta_metrics['ks'].statistic
        assert np.isclose(ks_stat, KSVAL)

        assert np.isclose(meta_metrics['outlier_rate'], OUTRATE)

    def test_pit_metric_small_eval_grid(self):
        """Test PIT metric warning message when number of pit samples is smaller than the evaluation grid"""
        with self.assertLogs(level='WARNING') as log:
            quant_grid = np.linspace(0, 1, 1000)
            _ = PIT(self.grid_ens, self.true_zs, quant_grid)
            self.assertIn('Number of pit samples is smaller', log.output[0])



"""This module implements a PDT distribution sub-class using interpolated quantiles
"""
import logging
import sys
import numpy as np

from scipy.stats import rv_continuous

from qp.pdf_gen import Pdf_rows_gen

from qp.conversion_funcs import extract_quantiles
from qp.plotting import get_axes_and_xlims, plot_pdf_quantiles_on_axes
from qp.utils import interpolate_multi_x_y, interpolate_x_multi_y, reshape_to_pdf_size
from qp.test_data import QUANTS, QLOCS, TEST_XVALS
from qp.factory import add_class

from qp.quantile_pdf_constructors import \
    AbstractQuantilePdfConstructor, \
    CdfSplineDerivative, \
    DualSplineAverage, \
    PiecewiseConstant, \
    PiecewiseLinear


epsilon = sys.float_info.epsilon

def pad_quantiles(quants, locs):
    """Pad the quantiles and locations used to build a quantile representation

    Paramters
    ---------
    quants : array_like
        The quantiles used to build the CDF
    locs : array_like
        The locations at which those quantiles are reached

    Returns
    -------
    quants : array_like
        The quantiles used to build the CDF
    locs : array_like
        The locations at which those quantiles are reached
    """
    n_out = n_vals = quants.size
    if quants[0] > sys.float_info.epsilon:
        offset_lo = 1
        pad_lo = True
        n_out += 1
    else:
        offset_lo = 0
        pad_lo = False
    if quants[-1] < 1.:
        pad_hi = True
        n_out += 1
    else:
        pad_hi = False
    if n_out == n_vals:
        return quants, locs
    quants_out = np.zeros((n_out), quants.dtype)
    locs_out = np.zeros((locs.shape[0], n_out), quants.dtype)
    quants_out[offset_lo:n_vals+offset_lo] = quants
    locs_out[:,offset_lo:n_vals+offset_lo] = locs
    if pad_lo:
        locs_out[:, 0] = locs[:, 0] - quants[0] * (locs[:, 1] - locs[:, 0]) / (quants[1] - quants[0])

    if pad_hi:
        quants_out[-1] = 1.
        locs_out[:, -1] = locs[:, -1] - (1. - quants[-1]) * (locs[:, -2] - locs[:, -1]) / (quants[-1] - quants[-2])

    return quants_out, locs_out

DEFAULT_PDF_CONSTRUCTOR = 'piecewise_linear'
PDF_CONSTRUCTORS = {
    'cdf_spline_derivative': CdfSplineDerivative,
    'dual_spline_average': DualSplineAverage,
    'piecewise_linear': PiecewiseLinear,
    'piecewise_constant': PiecewiseConstant,
}

class quant_gen(Pdf_rows_gen):
    """Quantile based distribution, where the PDF is defined piecewise from the quantiles

    Notes
    -----
    This implements a CDF by interpolating a set of quantile values

    It simply takes a set of x and y values and uses `scipy.interpolate.interp1d` to
    build the CDF
    """
    # pylint: disable=protected-access

    name = 'quant'
    version = 0

    _support_mask = rv_continuous._support_mask

    def __init__(self, quants, locs, *args, **kwargs):
        """
        Create a new distribution using the given values

        Parameters
        ----------
        quants : array_like
           The quantiles used to build the CDF
        locs : array_like
           The locations at which those quantiles are reached
        """

        self._xmin = np.min(locs)
        self._xmax = np.max(locs)

        locs_2d = reshape_to_pdf_size(locs, -1)
        self._check_input = kwargs.pop('check_input', True)
        if self._check_input:
            quants, locs_2d = pad_quantiles(quants, locs_2d)

        self._quants = np.asarray(quants)
        self._nquants = self._quants.size
        if locs_2d.shape[-1] != self._nquants:  # pragma: no cover
            raise ValueError("Number of locations (%i) != number of quantile values (%i)" % (self._nquants, locs_2d.shape[-1]))
        self._locs = locs_2d

        self._pdf_constructor_name = str(kwargs.pop('pdf_constructor_name', DEFAULT_PDF_CONSTRUCTOR))
        self._pdf_constructor = None
        self._instantiate_pdf_constructor()

        kwargs['shape'] = locs.shape[:-1]
        super().__init__(*args, **kwargs)

        self._addmetadata('quants', self._quants)
        self._addmetadata('pdf_constructor_name', self._pdf_constructor_name)
        self._addmetadata('check_input', self._check_input)
        self._addobjdata('locs', self._locs)

    @property
    def quants(self):
        """Return quantiles used to build the CDF"""
        return self._quants

    @property
    def locs(self):
        """Return the locations at which those quantiles are reached"""
        return self._locs

    @property
    def pdf_constructor_name(self):
        """Returns the name of the current pdf constructor. Matches a key in
        the PDF_CONSTRUCTORS dictionary."""
        return self._pdf_constructor_name

    @pdf_constructor_name.setter
    def pdf_constructor_name(self, value: str):
        """Allows users to specify a different interpolator without having to recreate
        the ensemble.

        Parameters
        ----------
        value : str
            One of the supported interpolators. See PDF_CONSTRUCTORS
            dictionary for supported interpolators.

        Raises
        ------
        ValueError
            If the value provided isn't a key in PDF_CONSTRUCTORS, raise
            a value error.
        """
        if value not in PDF_CONSTRUCTORS:
            raise ValueError(f"Unknown interpolator provided: '{value}'. Allowed interpolators are {list(PDF_CONSTRUCTORS.keys())}")

        if value is self._pdf_constructor_name:
            logging.warning("Already using interpolator: '%s'.", value)
            return

        self._pdf_constructor_name = value
        self._instantiate_pdf_constructor()
        self._addmetadata('pdf_constructor_name', self._pdf_constructor_name)

    @property
    def pdf_constructor(self) -> AbstractQuantilePdfConstructor:
        """Returns the current PDF constructor, and allows the user to interact
        with its methods.

        Returns
        -------
        AbstractQuantilePdfConstructor
            Abstract base class of the active concrete PDF constructor.
        """
        return self._pdf_constructor

    def _instantiate_pdf_constructor(self):
        self._pdf_constructor = PDF_CONSTRUCTORS[self._pdf_constructor_name](
            self._quants, self._locs)


    def _pdf(self, x, *args):
        # We're not requiring that the output be normalized!
        # `util.normalize_interp1d` addresses _one_ of the ways that a reconstruction
        # can be bad, but not all. It should be replaced with a more comprehensive
        # normalization function.
        # See qp issue #147
        row = args[0]
        return self._pdf_constructor.construct_pdf(x, row)

    def _cdf(self, x, row):
        # pylint: disable=arguments-differ
        return interpolate_multi_x_y(x, row, self._locs, self._quants,
                                     bounds_error=False, fill_value=(0., 1), kind="quadratic").ravel()

    def _ppf(self, x, row):
        # pylint: disable=arguments-differ
        return interpolate_x_multi_y(x, row, self._quants, self._locs,
                                     bounds_error=False, fill_value=(self._xmin, self._xmax), kind="quadratic").ravel()

    def _updated_ctor_param(self):
        """
        Set the quants and locs as additional constructor arguments
        """
        dct = super()._updated_ctor_param()
        dct['quants'] = self._quants
        dct['locs'] = self._locs
        dct['pdf_constructor_name'] = self._pdf_constructor_name
        dct['check_input'] = self._check_input
        return dct

    @classmethod
    def get_allocation_kwds(cls, npdf, **kwargs):
        """Return kwds necessary to create 'empty' hdf5 file with npdf entries
        for iterative writeout.  We only need to allocate the objdata columns, as
        the metadata can be written when we finalize the file.
        """
        try:
            quants = kwargs['quants']
        except ValueError: #pragma: no cover
            print("required argument 'quants' not included in kwargs")
        nquants = np.shape(quants)[-1]
        return dict(locs=((npdf, nquants), 'f4'))

    @classmethod
    def plot_native(cls, pdf, **kwargs):
        """Plot the PDF in a way that is particular to this type of distibution

        For a quantile this shows the quantiles points
        """
        axes, xlim, kw = get_axes_and_xlims(**kwargs)
        xvals = np.linspace(xlim[0], xlim[1], kw.pop('npts', 101))
        locs = np.squeeze(pdf.dist.locs[pdf.kwds['row']])
        quants = np.squeeze(pdf.dist.quants)
        yvals = np.squeeze(pdf.pdf(xvals))
        return plot_pdf_quantiles_on_axes(axes, xvals, yvals, quantiles=(quants, locs), **kw)

    @classmethod
    def add_mappings(cls):
        """
        Add this classes mappings to the conversion dictionary
        """
        cls._add_creation_method(cls.create, None)
        cls._add_extraction_method(extract_quantiles, None)


quant = quant_gen.create

quant_gen.test_data = dict(quant=dict(gen_func=quant, ctor_data=dict(quants=QUANTS, locs=QLOCS),\
                                          convert_data=dict(quants=QUANTS), test_xvals=TEST_XVALS))

add_class(quant_gen)



from .abstract_pdf_constructor import AbstractQuantilePdfConstructor
from .cdf_spline_derivative import CdfSplineDerivative
from .dual_spline_average import DualSplineAverage
from .piecewise_constant import PiecewiseConstant
from .piecewise_linear import PiecewiseLinear



"""
Unit tests for PDF class
"""

import unittest
import qp
import qp.metrics
import numpy as np

from qp import test_funcs

import time

class ScipyVectorizationTests(unittest.TestCase):
    """ Tests for the metrics """

    @classmethod
    def setUpClass(cls):
        """
        Make any objects that are used in multiple tests.
        """
        cls.random_state = 9999
        cls.rvs_size = 100
        t1 = time.perf_counter()
        # ! Consider changing this to the gamma distribution - norm is symmetric. trunc_norm is a good option too.
        cls.ens_n = test_funcs.build_ensemble(qp.stats.norm_gen.test_data['norm'])  #pylint: disable=no-member
        t2 = time.perf_counter()
        print(f'Setup time is {t2-t1}')
        print(f'Ensemble contains {cls.ens_n.npdf} distributions.')
        print(f'For each statistic {cls.rvs_size} random variables will be generated.')

    @classmethod
    def tearDownClass(cls):
        """ Clean up any mock data files created by the tests. """

    def test_ad_results(self):
        """This test compares the results of the Anderson-Darling evaluation against a known output"""
        expected_output = [1.41296516, 2.18126088, 0.59390624, 0.43942853, 0.67658563, 0.38376997,
        0.66460796, 2.72899571, 1.4542696, 1.34378282, 1.19221443]

        # Specify these here, so that changes in setupClass don't affect the results
        rvs_size = 100
        random_state = 9999

        test_output = qp.metrics.calculate_goodness_of_fit(
            self.ens_n,
            self.ens_n,
            fit_metric='ad',
            num_samples=rvs_size,
            _random_state=random_state
        )

        for test, expected in zip(test_output, expected_output):
            assert np.isclose(test, expected)

    def test_call_copied_method_ad_versus_single_distribution_ensemble(self):
        """For Anderson-Darling stat stat for ensemble vs. ensemble-with-1-distribution"""
        expected_output = [1.41296516e+00, 5.94997248e+01, 1.76866646e+02, 3.76671747e+02,
        6.17435009e+02, 9.61477755e+02, 1.24161826e+03, 2.33034254e+03, 1.87698332e+03,
        2.44635392e+03, 4.10601407e+03]

        # Specify these here, so that changes in setupClass don't affect the results
        rvs_size = 100
        random_state = 9999

        test_output = qp.metrics.calculate_goodness_of_fit(
            self.ens_n,
            self.ens_n[0],
            fit_metric='ad',
            num_samples=rvs_size,
            _random_state=random_state
        )

        for test, expected in zip(test_output, expected_output):
            assert np.isclose(test, expected)

    def test_cvm_results(self):
        """This test compares the results of the Cramer-von Mises evaluation against a known output"""
        expected_output = [0.22630139, 0.27090351, 0.09884569, 0.04506071, 0.10731862, 0.05631584,
            0.07646541, 0.52543176, 0.23656815, 0.21094075, 0.22384716]

        # Specify these here, so that changes in setupClass don't affect the results
        rvs_size = 100
        random_state = 9999

        test_output = qp.metrics.calculate_goodness_of_fit(
            self.ens_n,
            self.ens_n,
            fit_metric='cvm',
            num_samples=rvs_size,
            _random_state=random_state
        )

        for test, expected in zip(test_output, expected_output):
            assert np.isclose(test, expected)

    def test_call_copied_method_cvm_versus_single_distribution_ensemble(self):
        """For Cramer von Mises stat stat for ensemble vs. ensemble-with-1-distribution"""
        expected_output = [0.22630139, 4.90875135, 12.18815523, 17.88155095, 21.76168586, 24.17512056,
        22.69008766, 28.436041, 23.2801267, 23.87779491, 28.99530345]

        # Specify these here, so that changes in setupClass don't affect the results
        rvs_size = 100
        random_state = 9999

        test_output = qp.metrics.calculate_goodness_of_fit(
            self.ens_n,
            self.ens_n[0],
            fit_metric='cvm',
            num_samples=rvs_size,
            _random_state=random_state
        )

        for test, expected in zip(test_output, expected_output):
            assert np.isclose(test, expected)

    def test_ks_results(self):
        """This test compares the results of the Kolmogorov-Smirnov evaluation against a known output"""
        expected_output = [0.11739344, 0.10533858, 0.07950318, 0.06216017, 0.08471774, 0.05786293,
        0.08342168, 0.14441362, 0.09896913, 0.11243258, 0.09189444]

        # Specify these here, so that changes in setupClass don't affect the results
        rvs_size = 100
        random_state = 9999

        test_output = qp.metrics.calculate_goodness_of_fit(
            self.ens_n,
            self.ens_n,
            fit_metric='ks',
            num_samples=rvs_size,
            _random_state=random_state
        )

        for test, expected in zip(test_output, expected_output):
            assert np.isclose(test, expected)

    def test_call_copied_method_ks_versus_single_distribution_ensemble(self):
        """Test Kolmogorov-Smirnov stat for ensemble vs. ensemble-with-1-distribution"""
        expected_output = [0.11739344, 0.38809819, 0.51939786, 0.70188153, 0.76502763, 0.80196938,
        0.79053856, 0.8996195, 0.80379571, 0.82881527, 0.9095734]

        # Specify these here, so that changes in setupClass don't affect the results
        rvs_size = 100
        random_state = 9999

        test_output = qp.metrics.calculate_goodness_of_fit(
            self.ens_n,
            self.ens_n[0],
            fit_metric='ks',
            num_samples=rvs_size,
            _random_state=random_state
        )

        for test, expected in zip(test_output, expected_output):
            assert np.isclose(test, expected)
